# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_formats.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/ydb_formats.proto',
  package='Ydb.Formats',
  syntax='proto3',
  serialized_options=b'\n\020tech.ydb.formatsZ:github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Formats\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18protos/ydb_formats.proto\x12\x0bYdb.Formats\"$\n\x12\x41rrowBatchSettings\x12\x0e\n\x06schema\x18\x01 \x01(\x0c\"W\n\x0b\x43svSettings\x12\x11\n\tskip_rows\x18\x01 \x01(\r\x12\x11\n\tdelimiter\x18\x02 \x01(\x0c\x12\x12\n\nnull_value\x18\x03 \x01(\x0c\x12\x0e\n\x06header\x18\x04 \x01(\x08\x42Q\n\x10tech.ydb.formatsZ:github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Formats\xf8\x01\x01\x62\x06proto3'
)




_ARROWBATCHSETTINGS = _descriptor.Descriptor(
  name='ArrowBatchSettings',
  full_name='Ydb.Formats.ArrowBatchSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='Ydb.Formats.ArrowBatchSettings.schema', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=77,
)


_CSVSETTINGS = _descriptor.Descriptor(
  name='CsvSettings',
  full_name='Ydb.Formats.CsvSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='skip_rows', full_name='Ydb.Formats.CsvSettings.skip_rows', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delimiter', full_name='Ydb.Formats.CsvSettings.delimiter', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_value', full_name='Ydb.Formats.CsvSettings.null_value', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='header', full_name='Ydb.Formats.CsvSettings.header', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=79,
  serialized_end=166,
)

DESCRIPTOR.message_types_by_name['ArrowBatchSettings'] = _ARROWBATCHSETTINGS
DESCRIPTOR.message_types_by_name['CsvSettings'] = _CSVSETTINGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ArrowBatchSettings = _reflection.GeneratedProtocolMessageType('ArrowBatchSettings', (_message.Message,), {
  'DESCRIPTOR' : _ARROWBATCHSETTINGS,
  '__module__' : 'protos.ydb_formats_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Formats.ArrowBatchSettings)
  })
_sym_db.RegisterMessage(ArrowBatchSettings)

CsvSettings = _reflection.GeneratedProtocolMessageType('CsvSettings', (_message.Message,), {
  'DESCRIPTOR' : _CSVSETTINGS,
  '__module__' : 'protos.ydb_formats_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Formats.CsvSettings)
  })
_sym_db.RegisterMessage(CsvSettings)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from ydb import *  # noqa
import sys
import six
import warnings

warnings.warn("module kikimr.public.sdk.python.client is deprecated. please use ydb instead")


for name, module in six.iteritems(sys.modules.copy()):
    if not name.startswith("ydb"):
        continue

    if name.startswith("ydb.public"):
        continue

    module_import_path = name.split('.')
    if len(module_import_path) < 2:
        continue

    sys.modules['kikimr.public.sdk.python.client.' + '.'.join(module_import_path[1:])] = module



# -*- coding: utf-8 -*-
import abc
import ydb
from abc import abstractmethod
import logging
import time
import random
import enum

import six
from . import (
    issues,
    convert,
    settings as settings_impl,
    scheme,
    types,
    _utilities,
    _apis,
    _sp_impl,
    _session_impl,
    _tx_ctx_impl,
    tracing,
)
from ._errors import check_retriable_error

try:
    from . import interceptor
except ImportError:
    interceptor = None

logger = logging.getLogger(__name__)

##################################################################
# A deprecated aliases in case when direct import has been used  #
##################################################################
SessionPoolEmpty = issues.SessionPoolEmpty
DataQuery = types.DataQuery


class DescribeTableSettings(settings_impl.BaseRequestSettings):
    def __init__(self):
        super(DescribeTableSettings, self).__init__()
        self.include_shard_key_bounds = False
        self.include_table_stats = False

    def with_include_shard_key_bounds(self, value):
        self.include_shard_key_bounds = value
        return self

    def with_include_table_stats(self, value):
        self.include_table_stats = value
        return self


class ExecDataQuerySettings(settings_impl.BaseRequestSettings):
    def __init__(self):
        super(ExecDataQuerySettings, self).__init__()
        self.keep_in_cache = True

    def with_keep_in_cache(self, value):
        self.keep_in_cache = value
        return self


class KeyBound(object):
    __slots__ = ("_equal", "value", "type")

    def __init__(self, key_value, key_type=None, inclusive=False):
        """
        Represents key bound.
        :param key_value: An iterable with key values
        :param key_type: A type of key
        :param inclusive: A flag that indicates bound includes key provided in the value.
        """

        try:
            iter(key_value)
        except TypeError:
            assert False, "value must be iterable!"

        if isinstance(key_type, types.TupleType):
            key_type = key_type.proto

        self._equal = inclusive
        self.value = key_value
        self.type = key_type

    def is_inclusive(self):
        return self._equal

    def is_exclusive(self):
        return not self._equal

    def __str__(self):
        if self._equal:
            return "InclusiveKeyBound(Tuple%s)" % str(self.value)
        return "ExclusiveKeyBound(Tuple%s)" % str(self.value)

    @classmethod
    def inclusive(cls, key_value, key_type):
        return cls(key_value, key_type, True)

    @classmethod
    def exclusive(cls, key_value, key_type):
        return cls(key_value, key_type, False)


class KeyRange(object):
    __slots__ = ("from_bound", "to_bound")

    def __init__(self, from_bound, to_bound):
        self.from_bound = from_bound
        self.to_bound = to_bound

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "KeyRange(%s, %s)" % (str(self.from_bound), str(self.to_bound))


class Column(object):
    def __init__(self, name, type, family=None):
        self._name = name
        self._type = type
        self.family = family

    def __eq__(self, other):
        return self.name == other.name and self._type.item == other.type.item

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    def with_family(self, family):
        self.family = family
        return self

    @property
    def type_pb(self):
        try:
            return self._type.proto
        except Exception:
            return self._type


@enum.unique
class FeatureFlag(enum.IntEnum):
    UNSPECIFIED = 0
    ENABLED = 1
    DISABLED = 2


@enum.unique
class AutoPartitioningPolicy(enum.IntEnum):
    AUTO_PARTITIONING_POLICY_UNSPECIFIED = 0
    DISABLED = 1
    AUTO_SPLIT = 2
    AUTO_SPLIT_MERGE = 3


@enum.unique
class IndexStatus(enum.IntEnum):
    INDEX_STATUS_UNSPECIFIED = 0
    READY = 1
    BUILDING = 2


class CachingPolicy(object):
    def __init__(self):
        self._pb = _apis.ydb_table.CachingPolicy()
        self.preset_name = None

    def with_preset_name(self, preset_name):
        self._pb.preset_name = preset_name
        self.preset_name = preset_name
        return self

    def to_pb(self):
        return self._pb


class ExecutionPolicy(object):
    def __init__(self):
        self._pb = _apis.ydb_table.ExecutionPolicy()
        self.preset_name = None

    def with_preset_name(self, preset_name):
        self._pb.preset_name = preset_name
        self.preset_name = preset_name
        return self

    def to_pb(self):
        return self._pb


class CompactionPolicy(object):
    def __init__(self):
        self._pb = _apis.ydb_table.CompactionPolicy()
        self.preset_name = None

    def with_preset_name(self, preset_name):
        self._pb.preset_name = preset_name
        self.preset_name = preset_name
        return self

    def to_pb(self):
        return self._pb


class SplitPoint(object):
    def __init__(self, *args):
        self._value = tuple(args)

    @property
    def value(self):
        return self._value


class ExplicitPartitions(object):
    def __init__(self, split_points):
        self.split_points = split_points


class PartitioningPolicy(object):
    def __init__(self):
        self._pb = _apis.ydb_table.PartitioningPolicy()
        self.preset_name = None
        self.uniform_partitions = None
        self.auto_partitioning = None
        self.explicit_partitions = None

    def with_preset_name(self, preset_name):
        self._pb.preset_name = preset_name
        self.preset_name = preset_name
        return self

    def with_uniform_partitions(self, uniform_partitions):
        self._pb.uniform_partitions = uniform_partitions
        self.uniform_partitions = uniform_partitions
        return self

    def with_explicit_partitions(self, explicit_partitions):
        self.explicit_partitions = explicit_partitions
        return self

    def with_auto_partitioning(self, auto_partitioning):
        self._pb.auto_partitioning = auto_partitioning
        self.auto_partitioning = auto_partitioning
        return self

    def to_pb(self, table_description):
        if self.explicit_partitions is not None:
            column_types = {}
            pk = set(table_description.primary_key)
            for column in table_description.columns:
                if column.name in pk:
                    column_types[column.name] = column.type

            for split_point in self.explicit_partitions.split_points:
                typed_value = self._pb.explicit_partitions.split_points.add()
                split_point_type = types.TupleType()
                prefix_size = len(split_point.value)
                for pl_el_id, pk_name in enumerate(table_description.primary_key):
                    if pl_el_id >= prefix_size:
                        break

                    split_point_type.add_element(column_types[pk_name])

                typed_value.type.MergeFrom(split_point_type.proto)
                typed_value.value.MergeFrom(
                    convert.from_native_value(split_point_type.proto, split_point.value)
                )

        return self._pb


class TableIndex(object):
    def __init__(self, name):
        self._pb = _apis.ydb_table.TableIndex()
        self._pb.name = name
        self.name = name
        self.index_columns = []
        # output only.
        self.status = None

    def with_global_index(self):
        self._pb.global_index.SetInParent()
        return self

    def with_index_columns(self, *columns):
        for column in columns:
            self._pb.index_columns.append(column)
            self.index_columns.append(column)
        return self

    def to_pb(self):
        return self._pb


class ReplicationPolicy(object):
    def __init__(self):
        self._pb = _apis.ydb_table.ReplicationPolicy()
        self.preset_name = None
        self.replicas_count = None
        self.allow_promotion = None
        self.create_per_availability_zone = None

    def with_preset_name(self, preset_name):
        self._pb.preset_name = preset_name
        self.preset_name = preset_name
        return self

    def with_replicas_count(self, replicas_count):
        self._pb.replicas_count = replicas_count
        self.replicas_count = replicas_count
        return self

    def with_create_per_availability_zone(self, create_per_availability_zone):
        self._pb.create_per_availability_zone = create_per_availability_zone
        self.create_per_availability_zone = create_per_availability_zone
        return self

    def with_allow_promotion(self, allow_promotion):
        self._pb.allow_promotion = allow_promotion
        self.allow_promotion = allow_promotion
        return self

    def to_pb(self):
        return self._pb


class StoragePool(object):
    def __init__(self, media):
        self.media = media

    def to_pb(self):
        return _apis.ydb_table.StoragePool(media=self.media)


class StoragePolicy(object):
    def __init__(self):
        self._pb = _apis.ydb_table.StoragePolicy()
        self.preset_name = None
        self.syslog = None
        self.log = None
        self.data = None
        self.keep_in_memory = None
        self.external = None

    def with_preset_name(self, preset_name):
        self._pb.preset_name = preset_name
        self.preset_name = preset_name
        return self

    def with_syslog_storage_settings(self, syslog_settings):
        self._pb.syslog.MergeFrom(syslog_settings.to_pb())
        self.syslog = syslog_settings
        return self

    def with_log_storage_settings(self, log_settings):
        self._pb.log.MergeFrom(log_settings.to_pb())
        self.log = log_settings
        return self

    def with_data_storage_settings(self, data_settings):
        self._pb.data.MergeFrom(data_settings.to_pb())
        self.data = data_settings
        return self

    def with_external_storage_settings(self, external_settings):
        self._pb.external.MergeFrom(external_settings.to_pb())
        self.external = external_settings
        return self

    def with_keep_in_memory(self, keep_in_memory):
        self._pb.keep_in_memory = keep_in_memory
        self.keep_in_memory = keep_in_memory
        return self

    def to_pb(self):
        return self._pb


class TableProfile(object):
    def __init__(self):
        self.preset_name = None
        self.compaction_policy = None
        self.partitioning_policy = None
        self.storage_policy = None
        self.execution_policy = None
        self.replication_policy = None
        self.caching_policy = None

    def with_preset_name(self, preset_name):
        self.preset_name = preset_name
        return self

    def with_compaction_policy(self, compaction_policy):
        self.compaction_policy = compaction_policy
        return self

    def with_partitioning_policy(self, partitioning_policy):
        self.partitioning_policy = partitioning_policy
        return self

    def with_execution_policy(self, execution_policy):
        self.execution_policy = execution_policy
        return self

    def with_caching_policy(self, caching_policy):
        self.caching_policy = caching_policy
        return self

    def with_storage_policy(self, storage_policy):
        self.storage_policy = storage_policy
        return self

    def with_replication_policy(self, replication_policy):
        self.replication_policy = replication_policy
        return self

    def to_pb(self, table_description):
        pb = _apis.ydb_table.TableProfile()

        if self.preset_name is not None:
            pb.preset_name = self.preset_name

        if self.execution_policy is not None:
            pb.execution_policy.MergeFrom(self.execution_policy.to_pb())

        if self.storage_policy is not None:
            pb.storage_policy.MergeFrom(self.storage_policy.to_pb())

        if self.replication_policy is not None:
            pb.replication_policy.MergeFrom(self.replication_policy.to_pb())

        if self.caching_policy is not None:
            pb.caching_policy.MergeFrom(self.caching_policy.to_pb())

        if self.compaction_policy is not None:
            pb.compaction_policy.MergeFrom(self.compaction_policy.to_pb())

        if self.partitioning_policy is not None:
            pb.partitioning_policy.MergeFrom(
                self.partitioning_policy.to_pb(table_description)
            )

        return pb


class DateTypeColumnModeSettings(object):
    def __init__(self, column_name, expire_after_seconds=0):
        self.column_name = column_name
        self.expire_after_seconds = expire_after_seconds

    def to_pb(self):
        pb = _apis.ydb_table.DateTypeColumnModeSettings()

        pb.column_name = self.column_name
        pb.expire_after_seconds = self.expire_after_seconds

        return pb


@enum.unique
class ColumnUnit(enum.IntEnum):
    UNIT_UNSPECIFIED = 0
    UNIT_SECONDS = 1
    UNIT_MILLISECONDS = 2
    UNIT_MICROSECONDS = 3
    UNIT_NANOSECONDS = 4


class ValueSinceUnixEpochModeSettings(object):
    def __init__(self, column_name, column_unit, expire_after_seconds=0):
        self.column_name = column_name
        self.column_unit = column_unit
        self.expire_after_seconds = expire_after_seconds

    def to_pb(self):
        pb = _apis.ydb_table.ValueSinceUnixEpochModeSettings()

        pb.column_name = self.column_name
        pb.column_unit = self.column_unit
        pb.expire_after_seconds = self.expire_after_seconds

        return pb


class TtlSettings(object):
    def __init__(self):
        self.date_type_column = None
        self.value_since_unix_epoch = None

    def with_date_type_column(self, column_name, expire_after_seconds=0):
        self.date_type_column = DateTypeColumnModeSettings(
            column_name, expire_after_seconds
        )
        return self

    def with_value_since_unix_epoch(
        self, column_name, column_unit, expire_after_seconds=0
    ):
        self.value_since_unix_epoch = ValueSinceUnixEpochModeSettings(
            column_name, column_unit, expire_after_seconds
        )
        return self

    def to_pb(self):
        pb = _apis.ydb_table.TtlSettings()

        if self.date_type_column is not None:
            pb.date_type_column.MergeFrom(self.date_type_column.to_pb())
        elif self.value_since_unix_epoch is not None:
            pb.value_since_unix_epoch.MergeFrom(self.value_since_unix_epoch.to_pb())
        else:
            raise RuntimeError("Unspecified ttl settings mode")

        return pb


class TableStats(object):
    def __init__(self):
        self.partitions = None
        self.store_size = 0

    def with_store_size(self, store_size):
        self.store_size = store_size
        return self

    def with_partitions(self, partitions):
        self.partitions = partitions
        return self


class ReadReplicasSettings(object):
    def __init__(self):
        self.per_az_read_replicas_count = 0
        self.any_az_read_replicas_count = 0

    def with_any_az_read_replicas_count(self, any_az_read_replicas_count):
        self.any_az_read_replicas_count = any_az_read_replicas_count
        return self

    def with_per_az_read_replicas_count(self, per_az_read_replicas_count):
        self.per_az_read_replicas_count = per_az_read_replicas_count
        return self

    def to_pb(self):
        pb = _apis.ydb_table.ReadReplicasSettings()
        if self.per_az_read_replicas_count > 0:
            pb.per_az_read_replicas_count = self.per_az_read_replicas_count
        elif self.any_az_read_replicas_count > 0:
            pb.any_az_read_replicas_count = self.any_az_read_replicas_count
        return pb


class PartitioningSettings(object):
    def __init__(self):
        self.partitioning_by_size = 0
        self.partition_size_mb = 0
        self.partitioning_by_load = 0
        self.min_partitions_count = 0
        self.max_partitions_count = 0

    def with_max_partitions_count(self, max_partitions_count):
        self.max_partitions_count = max_partitions_count
        return self

    def with_min_partitions_count(self, min_partitions_count):
        self.min_partitions_count = min_partitions_count
        return self

    def with_partitioning_by_load(self, partitioning_by_load):
        self.partitioning_by_load = partitioning_by_load
        return self

    def with_partition_size_mb(self, partition_size_mb):
        self.partition_size_mb = partition_size_mb
        return self

    def with_partitioning_by_size(self, partitioning_by_size):
        self.partitioning_by_size = partitioning_by_size
        return self

    def to_pb(self):
        pb = _apis.ydb_table.PartitioningSettings()
        pb.partitioning_by_size = self.partitioning_by_size
        pb.partition_size_mb = self.partition_size_mb
        pb.partitioning_by_load = self.partitioning_by_load
        pb.min_partitions_count = self.min_partitions_count
        pb.max_partitions_count = self.max_partitions_count
        return pb


class StorageSettings(object):
    def __init__(self):
        self.tablet_commit_log0 = None
        self.tablet_commit_log1 = None
        self.external = None
        self.store_external_blobs = 0

    def with_store_external_blobs(self, store_external_blobs):
        self.store_external_blobs = store_external_blobs
        return self

    def with_external(self, external):
        self.external = external
        return self

    def with_tablet_commit_log1(self, tablet_commit_log1):
        self.tablet_commit_log1 = tablet_commit_log1
        return self

    def with_tablet_commit_log0(self, tablet_commit_log0):
        self.tablet_commit_log0 = tablet_commit_log0
        return self

    def to_pb(self):
        st = _apis.ydb_table.StorageSettings()
        st.store_external_blobs = self.store_external_blobs
        if self.external:
            st.external.MergeFrom(self.external.to_pb())
        if self.tablet_commit_log0:
            st.tablet_commit_log0.MergeFrom(self.tablet_commit_log0.to_pb())
        if self.tablet_commit_log1:
            st.tablet_commit_log1.MergeFrom(self.tablet_commit_log1.to_pb())
        return st


@enum.unique
class Compression(enum.IntEnum):
    UNSPECIFIED = 0
    NONE = 1
    LZ4 = 2


class ColumnFamily(object):
    def __init__(self):
        self.compression = 0
        self.name = None
        self.data = None
        self.keep_in_memory = 0

    def with_name(self, name):
        self.name = name
        return self

    def with_compression(self, compression):
        self.compression = compression
        return self

    def with_data(self, data):
        self.data = data
        return self

    def with_keep_in_memory(self, keep_in_memory):
        self.keep_in_memory = keep_in_memory
        return self

    def to_pb(self):
        cm = _apis.ydb_table.ColumnFamily()
        cm.keep_in_memory = self.keep_in_memory
        cm.compression = self.compression
        if self.name is not None:
            cm.name = self.name
        if self.data is not None:
            cm.data.MergeFrom(self.data.to_pb())
        return cm


class TableDescription(object):
    def __init__(self):
        self.columns = []
        self.primary_key = []
        self.profile = None
        self.indexes = []
        self.column_families = []
        self.ttl_settings = None
        self.attributes = {}
        self.uniform_partitions = 0
        self.partition_at_keys = None
        self.compaction_policy = None
        self.key_bloom_filter = 0
        self.read_replicas_settings = None
        self.partitioning_settings = None
        self.storage_settings = None

    def with_storage_settings(self, storage_settings):
        self.storage_settings = storage_settings
        return self

    def with_column(self, column):
        self.columns.append(column)
        return self

    def with_columns(self, *columns):
        for column in columns:
            self.with_column(column)
        return self

    def with_primary_key(self, key):
        self.primary_key.append(key)
        return self

    def with_primary_keys(self, *keys):
        for pk in keys:
            self.with_primary_key(pk)
        return self

    def with_column_family(self, column_family):
        self.column_families.append(column_family)
        return self

    def with_column_families(self, *column_families):
        for column_family in column_families:
            self.with_column_family(column_family)
        return self

    def with_indexes(self, *indexes):
        for index in indexes:
            self.with_index(index)
        return self

    def with_index(self, index):
        self.indexes.append(index)
        return self

    def with_profile(self, profile):
        self.profile = profile
        return self

    def with_ttl(self, ttl_settings):
        self.ttl_settings = ttl_settings
        return self

    def with_attributes(self, attributes):
        self.attributes = attributes
        return self

    def with_uniform_partitions(self, uniform_partitions):
        self.uniform_partitions = uniform_partitions
        return self

    def with_partition_at_keys(self, partition_at_keys):
        self.partition_at_keys = partition_at_keys
        return self

    def with_key_bloom_filter(self, key_bloom_filter):
        self.key_bloom_filter = key_bloom_filter
        return self

    def with_partitioning_settings(self, partitioning_settings):
        self.partitioning_settings = partitioning_settings
        return self

    def with_read_replicas_settings(self, read_replicas_settings):
        self.read_replicas_settings = read_replicas_settings
        return self

    def with_compaction_policy(self, compaction_policy):
        self.compaction_policy = compaction_policy
        return self


@six.add_metaclass(abc.ABCMeta)
class AbstractTransactionModeBuilder(object):
    @property
    @abc.abstractmethod
    def name(self):
        pass

    @property
    @abc.abstractmethod
    def settings(self):
        pass


class SnapshotReadOnly(AbstractTransactionModeBuilder):
    __slots__ = ("_pb", "_name")

    def __init__(self):
        self._pb = _apis.ydb_table.SnapshotModeSettings()
        self._name = "snapshot_read_only"

    @property
    def settings(self):
        return self._pb

    @property
    def name(self):
        return self._name


class SerializableReadWrite(AbstractTransactionModeBuilder):
    __slots__ = ("_pb", "_name")

    def __init__(self):
        self._name = "serializable_read_write"
        self._pb = _apis.ydb_table.SerializableModeSettings()

    @property
    def settings(self):
        return self._pb

    @property
    def name(self):
        return self._name


class OnlineReadOnly(AbstractTransactionModeBuilder):
    __slots__ = ("_pb", "_name")

    def __init__(self):
        self._pb = _apis.ydb_table.OnlineModeSettings()
        self._pb.allow_inconsistent_reads = False
        self._name = "online_read_only"

    def with_allow_inconsistent_reads(self):
        self._pb.allow_inconsistent_reads = True
        return self

    @property
    def settings(self):
        return self._pb

    @property
    def name(self):
        return self._name


class StaleReadOnly(AbstractTransactionModeBuilder):
    __slots__ = ("_pb", "_name")

    def __init__(self):
        self._pb = _apis.ydb_table.StaleModeSettings()
        self._name = "stale_read_only"

    @property
    def settings(self):
        return self._pb

    @property
    def name(self):
        return self._name


class BackoffSettings(object):
    def __init__(self, ceiling=6, slot_duration=0.001, uncertain_ratio=0.5):
        self.ceiling = ceiling
        self.slot_duration = slot_duration
        self.uncertain_ratio = uncertain_ratio

    def calc_timeout(self, retry_number):
        slots_count = 1 << min(retry_number, self.ceiling)
        max_duration_ms = slots_count * self.slot_duration * 1000.0
        # duration_ms = random.random() * max_duration_ms * uncertain_ratio) + max_duration_ms * (1 - uncertain_ratio)
        duration_ms = max_duration_ms * (
            random.random() * self.uncertain_ratio + 1.0 - self.uncertain_ratio
        )
        return duration_ms / 1000.0


class RetrySettings(object):
    def __init__(
        self,
        max_retries=10,
        max_session_acquire_timeout=None,
        on_ydb_error_callback=None,
        backoff_ceiling=6,
        backoff_slot_duration=1,
        get_session_client_timeout=5,
        fast_backoff_settings=None,
        slow_backoff_settings=None,
        idempotent=False,
    ):
        self.max_retries = max_retries
        self.max_session_acquire_timeout = max_session_acquire_timeout
        self.on_ydb_error_callback = (
            (lambda e: None) if on_ydb_error_callback is None else on_ydb_error_callback
        )
        self.fast_backoff = (
            BackoffSettings(10, 0.005)
            if fast_backoff_settings is None
            else fast_backoff_settings
        )
        self.slow_backoff = (
            BackoffSettings(backoff_ceiling, backoff_slot_duration)
            if slow_backoff_settings is None
            else slow_backoff_settings
        )
        self.retry_not_found = True
        self.idempotent = idempotent
        self.retry_internal_error = True
        self.unknown_error_handler = lambda e: None
        self.get_session_client_timeout = get_session_client_timeout
        if max_session_acquire_timeout is not None:
            self.get_session_client_timeout = min(
                self.max_session_acquire_timeout, self.get_session_client_timeout
            )

    def with_fast_backoff(self, backoff_settings):
        self.fast_backoff = backoff_settings
        return self

    def with_slow_backoff(self, backoff_settings):
        self.slow_backoff = backoff_settings
        return self


class YdbRetryOperationSleepOpt(object):
    def __init__(self, timeout):
        self.timeout = timeout

    def __eq__(self, other):
        return type(self) == type(other) and self.timeout == other.timeout

    def __repr__(self):
        return "YdbRetryOperationSleepOpt(%s)" % self.timeout


class YdbRetryOperationFinalResult(object):
    def __init__(self, result):
        self.result = result
        self.exc = None

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.result == other.result
            and self.exc == other.exc
        )

    def __repr__(self):
        return "YdbRetryOperationFinalResult(%s, exc=%s)" % (self.result, self.exc)

    def set_exception(self, exc):
        self.exc = exc


def retry_operation_impl(callee, retry_settings=None, *args, **kwargs):
    retry_settings = RetrySettings() if retry_settings is None else retry_settings
    status = None

    for attempt in six.moves.range(retry_settings.max_retries + 1):
        try:
            result = YdbRetryOperationFinalResult(callee(*args, **kwargs))
            yield result

            if result.exc is not None:
                raise result.exc

        except issues.Error as e:
            status = e
            retry_settings.on_ydb_error_callback(e)

            retriable_info = check_retriable_error(e, retry_settings, attempt)
            if not retriable_info.is_retriable:
                raise

            skip_yield_error_types = [
                issues.Aborted,
                issues.BadSession,
                issues.NotFound,
                issues.InternalError,
            ]

            yield_sleep = True
            for t in skip_yield_error_types:
                if isinstance(e, t):
                    yield_sleep = False

            if yield_sleep:
                yield YdbRetryOperationSleepOpt(retriable_info.sleep_timeout_seconds)

        except Exception as e:
            # you should provide your own handler you want
            retry_settings.unknown_error_handler(e)
            raise

    raise status


def retry_operation_sync(callee, retry_settings=None, *args, **kwargs):
    opt_generator = retry_operation_impl(callee, retry_settings, *args, **kwargs)
    for next_opt in opt_generator:
        if isinstance(next_opt, YdbRetryOperationSleepOpt):
            time.sleep(next_opt.timeout)
        else:
            return next_opt.result


class TableClientSettings(object):
    def __init__(self):
        self._client_query_cache_enabled = False
        self._native_datetime_in_result_sets = False
        self._native_date_in_result_sets = False
        self._make_result_sets_lazy = False
        self._native_json_in_result_sets = False
        self._native_interval_in_result_sets = False
        self._native_timestamp_in_result_sets = False

    def with_native_timestamp_in_result_sets(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._native_timestamp_in_result_sets = enabled
        return self

    def with_native_interval_in_result_sets(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._native_interval_in_result_sets = enabled
        return self

    def with_native_json_in_result_sets(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._native_json_in_result_sets = enabled
        return self

    def with_native_date_in_result_sets(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._native_date_in_result_sets = enabled
        return self

    def with_native_datetime_in_result_sets(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._native_datetime_in_result_sets = enabled
        return self

    def with_client_query_cache(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._client_query_cache_enabled = enabled
        return self

    def with_lazy_result_sets(self, enabled):
        # type:(bool) -> ydb.TableClientSettings
        self._make_result_sets_lazy = enabled
        return self


class ScanQueryResult(object):
    def __init__(self, result, table_client_settings):
        self._result = result
        self.query_stats = result.query_stats
        self.result_set = convert.ResultSet.from_message(
            self._result.result_set, table_client_settings
        )


@enum.unique
class QueryStatsCollectionMode(enum.IntEnum):
    NONE = _apis.ydb_table.QueryStatsCollection.Mode.STATS_COLLECTION_NONE
    BASIC = _apis.ydb_table.QueryStatsCollection.Mode.STATS_COLLECTION_BASIC
    FULL = _apis.ydb_table.QueryStatsCollection.Mode.STATS_COLLECTION_FULL


class ScanQuerySettings(settings_impl.BaseRequestSettings):
    def __init__(self):
        super(ScanQuerySettings, self).__init__()
        self.collect_stats = None

    def with_collect_stats(self, collect_stats_mode):
        self.collect_stats = collect_stats_mode
        return self


class ScanQuery(object):
    def __init__(self, yql_text, parameters_types):
        self.yql_text = yql_text
        self.parameters_types = parameters_types


def _wrap_scan_query_response(response, table_client_settings):
    issues._process_response(response)
    return ScanQueryResult(response.result, table_client_settings)


def _scan_query_request_factory(query, parameters=None, settings=None):
    if not isinstance(query, ScanQuery):
        query = ScanQuery(query, {})
    parameters = {} if parameters is None else parameters
    collect_stats = getattr(
        settings,
        "collect_stats",
        _apis.ydb_table.QueryStatsCollection.Mode.STATS_COLLECTION_NONE,
    )
    return _apis.ydb_table.ExecuteScanQueryRequest(
        mode=_apis.ydb_table.ExecuteScanQueryRequest.Mode.MODE_EXEC,
        query=_apis.ydb_table.Query(yql_text=query.yql_text),
        parameters=convert.parameters_to_pb(query.parameters_types, parameters),
        collect_stats=collect_stats,
    )


@six.add_metaclass(abc.ABCMeta)
class ISession:
    @abstractmethod
    def __init__(self, driver, table_client_settings):
        pass

    @abstractmethod
    def __lt__(self, other):
        pass

    @abstractmethod
    def __eq__(self, other):
        pass

    @property
    @abstractmethod
    def session_id(self):
        pass

    @abstractmethod
    def initialized(self):
        """
        Return True if session is successfully initialized with a session_id and False otherwise.
        """
        pass

    @abstractmethod
    def pending_query(self):
        pass

    @abstractmethod
    def reset(self):
        """
        Perform session state reset (that includes cleanup of the session_id, query cache, and etc.)
        """
        pass

    @abstractmethod
    def read_table(
        self,
        path,
        key_range=None,
        columns=(),
        ordered=False,
        row_limit=None,
        settings=None,
        use_snapshot=None,
    ):
        """
        Perform an read table request.

        :param path: A path to the table
        :param key_range: (optional) A KeyRange instance that describes a range to read. The KeyRange instance\
        should include from_bound and/or to_bound. Each of the bounds (if provided) should specify a value of the\
        key bound, and type of the key prefix. See an example above.
        :param columns: (optional) An iterable with table columns to read.
        :param ordered: (optional) A flag that indicates that result should be ordered.
        :param row_limit: (optional) A number of rows to read.
        :param settings: Request settings

        :return: SyncResponseIterator instance
        """
        pass

    @abstractmethod
    def keep_alive(self, settings=None):
        pass

    @abstractmethod
    def create(self, settings=None):
        pass

    @abstractmethod
    def delete(self, settings=None):
        pass

    @abstractmethod
    def execute_scheme(self, yql_text, settings=None):
        pass

    @abstractmethod
    def transaction(self, tx_mode=None):
        pass

    @abstractmethod
    def has_prepared(self, query):
        pass

    @abstractmethod
    def prepare(self, query, settings=None):
        pass

    @abstractmethod
    def explain(self, yql_text, settings=None):
        """
        Expiremental API.

        :param yql_text:
        :param settings:

        :return:
        """
        pass

    @abstractmethod
    def create_table(self, path, table_description, settings=None):
        """
        Create a YDB table.

        :param path: A table path
        :param table_description: A description of table to create. An instance TableDescription
        :param settings: An instance of BaseRequestSettings that describes how rpc should invoked.

        :return: A description of created scheme entry or error otherwise.
        """
        pass

    @abstractmethod
    def drop_table(self, path, settings=None):
        pass

    @abstractmethod
    def alter_table(
        self,
        path,
        add_columns=None,
        drop_columns=None,
        settings=None,
        alter_attributes=None,
        add_indexes=None,
        drop_indexes=None,
        set_ttl_settings=None,
        drop_ttl_settings=None,
        add_column_families=None,
        alter_column_families=None,
        alter_storage_settings=None,
        set_compaction_policy=None,
        alter_partitioning_settings=None,
        set_key_bloom_filter=None,
        set_read_replicas_settings=None,
    ):
        pass

    @abstractmethod
    def copy_table(self, source_path, destination_path, settings=None):
        pass

    @abstractmethod
    def copy_tables(self, source_destination_pairs, settings=None):
        pass

    def describe_table(self, path, settings=None):
        """
        Returns a description of the table by provided path

        :param path: A table path
        :param settings: A request settings

        :return: Description of a table
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class ITableClient:
    def __init__(self, driver, table_client_settings=None):
        pass

    @abstractmethod
    def session(self):
        pass

    @abstractmethod
    def scan_query(self, query, parameters=None, settings=None):
        pass

    @abstractmethod
    def bulk_upsert(self, table_path, rows, column_types, settings=None):
        """
        Bulk upsert data

        :param table_path: A table path.
        :param rows: A list of structures.
        :param column_types: Bulk upsert column types.

        """
        pass


class BaseTableClient(ITableClient):
    def __init__(self, driver, table_client_settings=None):
        # type:(ydb.Driver, ydb.TableClientSettings) -> None
        self._driver = driver
        self._table_client_settings = (
            TableClientSettings()
            if table_client_settings is None
            else table_client_settings
        )

    def session(self):
        # type: () -> ydb.Session
        return Session(self._driver, self._table_client_settings)

    def scan_query(self, query, parameters=None, settings=None):
        # type: (ydb.ScanQuery, tuple, ydb.BaseRequestSettings) -> ydb.SyncResponseIterator
        request = _scan_query_request_factory(query, parameters, settings)
        stream_it = self._driver(
            request,
            _apis.TableService.Stub,
            _apis.TableService.StreamExecuteScanQuery,
            settings=settings,
        )
        return _utilities.SyncResponseIterator(
            stream_it,
            lambda resp: _wrap_scan_query_response(resp, self._table_client_settings),
        )

    def bulk_upsert(self, table_path, rows, column_types, settings=None):
        # type: (str, list, ydb.AbstractTypeBuilder | ydb.PrimitiveType, ydb.BaseRequestSettings) -> None
        """
        Bulk upsert data

        :param table_path: A table path.
        :param rows: A list of structures.
        :param column_types: Bulk upsert column types.

        """
        return self._driver(
            _session_impl.bulk_upsert_request_factory(table_path, rows, column_types),
            _apis.TableService.Stub,
            _apis.TableService.BulkUpsert,
            _session_impl.wrap_operation_bulk_upsert,
            settings,
            (),
        )


class TableClient(BaseTableClient):
    def async_scan_query(self, query, parameters=None, settings=None):
        # type: (ydb.ScanQuery, tuple, ydb.BaseRequestSettings) -> ydb.AsyncResponseIterator
        request = _scan_query_request_factory(query, parameters, settings)
        stream_it = self._driver(
            request,
            _apis.TableService.Stub,
            _apis.TableService.StreamExecuteScanQuery,
            settings=settings,
        )
        return _utilities.AsyncResponseIterator(
            stream_it,
            lambda resp: _wrap_scan_query_response(resp, self._table_client_settings),
        )

    @_utilities.wrap_async_call_exceptions
    def async_bulk_upsert(self, table_path, rows, column_types, settings=None):
        # type: (str, list, ydb.AbstractTypeBuilder | ydb.PrimitiveType, ydb.BaseRequestSettings) -> None
        return self._driver.future(
            _session_impl.bulk_upsert_request_factory(table_path, rows, column_types),
            _apis.TableService.Stub,
            _apis.TableService.BulkUpsert,
            _session_impl.wrap_operation_bulk_upsert,
            settings,
            (),
        )


def _make_index_description(index):
    result = TableIndex(index.name).with_index_columns(
        *tuple(col for col in index.index_columns)
    )
    result.status = IndexStatus(index.status)
    return result


class TableSchemeEntry(scheme.SchemeEntry):
    def __init__(
        self,
        name,
        owner,
        type,
        effective_permissions,
        permissions,
        size_bytes,
        columns,
        primary_key,
        shard_key_bounds,
        indexes,
        table_stats,
        ttl_settings,
        attributes,
        partitioning_settings,
        column_families,
        key_bloom_filter,
        read_replicas_settings,
        storage_settings,
        *args,
        **kwargs
    ):

        super(TableSchemeEntry, self).__init__(
            name,
            owner,
            type,
            effective_permissions,
            permissions,
            size_bytes,
            *args,
            **kwargs
        )
        self.primary_key = [pk for pk in primary_key]
        self.columns = [
            Column(column.name, convert.type_to_native(column.type), column.family)
            for column in columns
        ]
        self.indexes = [_make_index_description(index) for index in indexes]
        self.shard_key_ranges = []
        self.column_families = []
        self.key_bloom_filter = FeatureFlag(key_bloom_filter)
        left_key_bound = None
        for column_family in column_families:
            self.column_families.append(
                ColumnFamily()
                .with_name(column_family.name)
                .with_keep_in_memory(FeatureFlag(column_family.keep_in_memory))
                .with_compression(Compression(column_family.compression))
            )

            if column_family.HasField("data"):
                self.column_families[-1].with_data(
                    StoragePool(column_family.data.media)
                )

        for shard_key_bound in shard_key_bounds:
            # for next key range
            key_bound_type = shard_key_bound.type
            current_bound = convert.to_native_value(shard_key_bound)
            self.shard_key_ranges.append(
                KeyRange(
                    None
                    if left_key_bound is None
                    else KeyBound.inclusive(left_key_bound, key_bound_type),
                    KeyBound.exclusive(current_bound, key_bound_type),
                )
            )
            left_key_bound = current_bound

            assert isinstance(left_key_bound, tuple)

        if len(shard_key_bounds) > 0:
            self.shard_key_ranges.append(
                KeyRange(
                    KeyBound.inclusive(left_key_bound, shard_key_bounds[-1].type),
                    None,
                )
            )

        else:
            self.shard_key_ranges.append(KeyRange(None, None))

        self.read_replicas_settings = None
        if read_replicas_settings is not None:
            self.read_replicas_settings = ReadReplicasSettings()
            for field in ("per_az_read_replicas_count", "any_az_read_replicas_count"):
                if read_replicas_settings.WhichOneof("settings") == field:
                    setattr(
                        self.read_replicas_settings,
                        field,
                        getattr(read_replicas_settings, field),
                    )

        self.storage_settings = None
        if storage_settings is not None:
            self.storage_settings = StorageSettings()
            self.storage_settings.store_external_blobs = FeatureFlag(
                self.storage_settings.store_external_blobs
            )
            if storage_settings.HasField("tablet_commit_log0"):
                self.storage_settings.with_tablet_commit_log0(
                    StoragePool(storage_settings.tablet_commit_log0.media)
                )

            if storage_settings.HasField("tablet_commit_log1"):
                self.storage_settings.with_tablet_commit_log1(
                    StoragePool(storage_settings.tablet_commit_log1.media)
                )

            if storage_settings.HasField("external"):
                self.storage_settings.with_external(
                    StoragePool(storage_settings.external.media)
                )

        self.partitioning_settings = None
        if partitioning_settings is not None:
            self.partitioning_settings = PartitioningSettings()
            for field in (
                "partitioning_by_size",
                "partitioning_by_load",
                "partition_size_mb",
                "min_partitions_count",
                "max_partitions_count",
            ):
                setattr(
                    self.partitioning_settings,
                    field,
                    getattr(partitioning_settings, field),
                )

        self.ttl_settings = None
        if ttl_settings is not None:
            if ttl_settings.HasField("date_type_column"):
                self.ttl_settings = TtlSettings().with_date_type_column(
                    ttl_settings.date_type_column.column_name,
                    ttl_settings.date_type_column.expire_after_seconds,
                )
            elif ttl_settings.HasField("value_since_unix_epoch"):
                self.ttl_settings = TtlSettings().with_value_since_unix_epoch(
                    ttl_settings.value_since_unix_epoch.column_name,
                    ColumnUnit(ttl_settings.value_since_unix_epoch.column_unit),
                    ttl_settings.value_since_unix_epoch.expire_after_seconds,
                )

        self.table_stats = None
        if table_stats is not None:
            self.table_stats = TableStats()
            if table_stats.partitions != 0:
                self.table_stats = self.table_stats.with_partitions(
                    table_stats.partitions
                )

            if table_stats.store_size != 0:
                self.table_stats = self.table_stats.with_store_size(
                    table_stats.store_size
                )

        self.attributes = attributes


class RenameItem:
    def __init__(self, source_path, destination_path, replace_destination=False):
        self._source_path = source_path
        self._destination_path = destination_path
        self._replace_destination = replace_destination

    @property
    def source_path(self):
        return self._source_path

    @property
    def destination_path(self):
        return self._destination_path

    @property
    def replace_destination(self):
        return self._replace_destination


class BaseSession(ISession):
    def __init__(self, driver, table_client_settings):
        self._driver = driver
        self._state = _session_impl.SessionState(table_client_settings)

    def __lt__(self, other):
        return self.session_id < other.session_id

    def __eq__(self, other):
        return self.session_id == other.session_id

    @property
    def session_id(self):
        """
        Return session_id.
        """
        return self._state.session_id

    def initialized(self):
        """
        Return True if session is successfully initialized with a session_id and False otherwise.
        """
        return self._state.session_id is not None

    def pending_query(self):
        return self._state.pending_query()

    def closing(self):
        """Returns True if session is closing."""
        return self._state.closing()

    def reset(self):
        """
        Perform session state reset (that includes cleanup of the session_id, query cache, and etc.)
        """
        return self._state.reset()

    def read_table(
        self,
        path,
        key_range=None,
        columns=(),
        ordered=False,
        row_limit=None,
        settings=None,
        use_snapshot=None,
    ):
        """
        Perform an read table request.

        :param path: A path to the table
        :param key_range: (optional) A KeyRange instance that describes a range to read. The KeyRange instance\
        should include from_bound and/or to_bound. Each of the bounds (if provided) should specify a value of the\
        key bound, and type of the key prefix. See an example above.
        :param columns: (optional) An iterable with table columns to read.
        :param ordered: (optional) A flag that indicates that result should be ordered.
        :param row_limit: (optional) A number of rows to read.

        :return: SyncResponseIterator instance
        """
        request = _session_impl.read_table_request_factory(
            self._state,
            path,
            key_range,
            columns,
            ordered,
            row_limit,
            use_snapshot=use_snapshot,
        )
        stream_it = self._driver(
            request,
            _apis.TableService.Stub,
            _apis.TableService.StreamReadTable,
            settings=settings,
        )
        return _utilities.SyncResponseIterator(
            stream_it, _session_impl.wrap_read_table_response
        )

    def keep_alive(self, settings=None):
        return self._driver(
            _session_impl.keep_alive_request_factory(self._state),
            _apis.TableService.Stub,
            _apis.TableService.KeepAlive,
            _session_impl.wrap_keep_alive_response,
            settings,
            (self._state, self),
            self._state.endpoint,
        )

    def create(self, settings=None):
        if self._state.session_id is not None:
            return self
        create_settings = settings_impl.BaseRequestSettings()
        if settings is not None:
            create_settings = settings.make_copy()
        create_settings = create_settings.with_header(
            "x-ydb-client-capabilities", "session-balancer"
        )
        return self._driver(
            _apis.ydb_table.CreateSessionRequest(),
            _apis.TableService.Stub,
            _apis.TableService.CreateSession,
            _session_impl.initialize_session,
            create_settings,
            (self._state, self),
            self._state.endpoint,
        )

    def delete(self, settings=None):
        return self._driver(
            self._state.attach_request(_apis.ydb_table.DeleteSessionRequest()),
            _apis.TableService.Stub,
            _apis.TableService.DeleteSession,
            _session_impl.cleanup_session,
            settings,
            (self._state, self),
            self._state.endpoint,
        )

    def execute_scheme(self, yql_text, settings=None):
        return self._driver(
            _session_impl.execute_scheme_request_factory(self._state, yql_text),
            _apis.TableService.Stub,
            _apis.TableService.ExecuteSchemeQuery,
            _session_impl.wrap_execute_scheme_result,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    def transaction(self, tx_mode=None):
        return TxContext(self._driver, self._state, self, tx_mode)

    def has_prepared(self, query):
        return query in self._state

    def prepare(self, query, settings=None):
        data_query, _ = self._state.lookup(query)
        if data_query is not None:
            return data_query
        return self._driver(
            _session_impl.prepare_request_factory(self._state, query),
            _apis.TableService.Stub,
            _apis.TableService.PrepareDataQuery,
            _session_impl.wrap_prepare_query_response,
            settings,
            (self._state, query),
            self._state.endpoint,
        )

    def explain(self, yql_text, settings=None):
        """
        Expiremental API.

        :param yql_text:
        :param settings:

        :return:
        """
        return self._driver(
            _session_impl.explain_data_query_request_factory(self._state, yql_text),
            _apis.TableService.Stub,
            _apis.TableService.ExplainDataQuery,
            _session_impl.wrap_explain_response,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    def create_table(self, path, table_description, settings=None):
        """
        Create a YDB table.

        :param path: A table path
        :param table_description: A description of table to create. An instance TableDescription
        :param settings: An instance of BaseRequestSettings that describes how rpc should invoked.

        :return: A description of created scheme entry or error otherwise.
        """
        return self._driver(
            _session_impl.create_table_request_factory(
                self._state, path, table_description
            ),
            _apis.TableService.Stub,
            _apis.TableService.CreateTable,
            _session_impl.wrap_operation,
            settings,
            (self._driver,),
            self._state.endpoint,
        )

    def drop_table(self, path, settings=None):
        return self._driver(
            self._state.attach_request(_apis.ydb_table.DropTableRequest(path=path)),
            _apis.TableService.Stub,
            _apis.TableService.DropTable,
            _session_impl.wrap_operation,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    def alter_table(
        self,
        path,
        add_columns=None,
        drop_columns=None,
        settings=None,
        alter_attributes=None,
        add_indexes=None,
        drop_indexes=None,
        set_ttl_settings=None,
        drop_ttl_settings=None,
        add_column_families=None,
        alter_column_families=None,
        alter_storage_settings=None,
        set_compaction_policy=None,
        alter_partitioning_settings=None,
        set_key_bloom_filter=None,
        set_read_replicas_settings=None,
    ):
        return self._driver(
            _session_impl.alter_table_request_factory(
                self._state,
                path,
                add_columns,
                drop_columns,
                alter_attributes,
                add_indexes,
                drop_indexes,
                set_ttl_settings,
                drop_ttl_settings,
                add_column_families,
                alter_column_families,
                alter_storage_settings,
                set_compaction_policy,
                alter_partitioning_settings,
                set_key_bloom_filter,
                set_read_replicas_settings,
            ),
            _apis.TableService.Stub,
            _apis.TableService.AlterTable,
            _session_impl.AlterTableOperation,
            settings,
            (self._driver,),
            self._state.endpoint,
        )

    def describe_table(self, path, settings=None):
        """
        Returns a description of the table by provided path

        :param path: A table path
        :param settings: A request settings

        :return: Description of a table
        """
        return self._driver(
            _session_impl.describe_table_request_factory(self._state, path, settings),
            _apis.TableService.Stub,
            _apis.TableService.DescribeTable,
            _session_impl.wrap_describe_table_response,
            settings,
            (self._state, TableSchemeEntry),
            self._state.endpoint,
        )

    def copy_table(self, source_path, destination_path, settings=None):
        return self.copy_tables([(source_path, destination_path)], settings=settings)

    def copy_tables(self, source_destination_pairs, settings=None):
        return self._driver(
            _session_impl.copy_tables_request_factory(
                self._state, source_destination_pairs
            ),
            _apis.TableService.Stub,
            _apis.TableService.CopyTables,
            _session_impl.wrap_operation,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    def rename_tables(self, rename_items, settings=None):
        return self._driver(
            _session_impl.rename_tables_request_factory(self._state, rename_items),
            _apis.TableService.Stub,
            _apis.TableService.RenameTables,
            _session_impl.wrap_operation,
            settings,
            (self._state,),
            self._state.endpoint,
        )


class Session(BaseSession):
    def async_read_table(
        self,
        path,
        key_range=None,
        columns=(),
        ordered=False,
        row_limit=None,
        settings=None,
        use_snapshot=None,
    ):
        """
        Perform an read table request.

        :param path: A path to the table
        :param key_range: (optional) A KeyRange instance that describes a range to read. The KeyRange instance\
        should include from_bound and/or to_bound. Each of the bounds (if provided) should specify a value of the\
        key bound, and type of the key prefix. See an example above.
        :param columns: (optional) An iterable with table columns to read.
        :param ordered: (optional) A flag that indicates that result should be ordered.
        :param row_limit: (optional) A number of rows to read.

        :return: AsyncResponseIterator instance
        """
        if interceptor is None:
            raise RuntimeError("Async read table is not available due to import issues")
        request = _session_impl.read_table_request_factory(
            self._state,
            path,
            key_range,
            columns,
            ordered,
            row_limit,
            use_snapshot=use_snapshot,
        )
        stream_it = self._driver(
            request,
            _apis.TableService.Stub,
            _apis.TableService.StreamReadTable,
            settings=settings,
        )
        return _utilities.AsyncResponseIterator(
            stream_it, _session_impl.wrap_read_table_response
        )

    @_utilities.wrap_async_call_exceptions
    def async_keep_alive(self, settings=None):
        return self._driver.future(
            _session_impl.keep_alive_request_factory(self._state),
            _apis.TableService.Stub,
            _apis.TableService.KeepAlive,
            _session_impl.wrap_keep_alive_response,
            settings,
            (self._state, self),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_create(self, settings=None):
        if self._state.session_id is not None:
            return _utilities.wrap_result_in_future(self)
        create_settings = settings_impl.BaseRequestSettings()
        if settings is not None:
            create_settings = settings.make_copy()
        create_settings = create_settings.with_header(
            "x-ydb-client-capabilities", "session-balancer"
        )
        return self._driver.future(
            _apis.ydb_table.CreateSessionRequest(),
            _apis.TableService.Stub,
            _apis.TableService.CreateSession,
            _session_impl.initialize_session,
            create_settings,
            (self._state, self),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_delete(self, settings=None):
        return self._driver.future(
            self._state.attach_request(_apis.ydb_table.DeleteSessionRequest()),
            _apis.TableService.Stub,
            _apis.TableService.DeleteSession,
            _session_impl.cleanup_session,
            settings,
            (self._state, self),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_execute_scheme(self, yql_text, settings=None):
        return self._driver.future(
            _session_impl.execute_scheme_request_factory(self._state, yql_text),
            _apis.TableService.Stub,
            _apis.TableService.ExecuteSchemeQuery,
            _session_impl.wrap_execute_scheme_result,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_prepare(self, query, settings=None):
        data_query, _ = self._state.lookup(query)
        if data_query is not None:
            return _utilities.wrap_result_in_future(data_query)
        return self._driver.future(
            _session_impl.prepare_request_factory(self._state, query),
            _apis.TableService.Stub,
            _apis.TableService.PrepareDataQuery,
            _session_impl.wrap_prepare_query_response,
            settings,
            (
                self._state,
                query,
            ),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_create_table(self, path, table_description, settings=None):
        return self._driver.future(
            _session_impl.create_table_request_factory(
                self._state, path, table_description
            ),
            _apis.TableService.Stub,
            _apis.TableService.CreateTable,
            _session_impl.wrap_operation,
            settings,
            (self._driver,),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_drop_table(self, path, settings=None):
        return self._driver.future(
            self._state.attach_request(_apis.ydb_table.DropTableRequest(path=path)),
            _apis.TableService.Stub,
            _apis.TableService.DropTable,
            _session_impl.wrap_operation,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_alter_table(
        self,
        path,
        add_columns=None,
        drop_columns=None,
        settings=None,
        alter_attributes=None,
        add_indexes=None,
        drop_indexes=None,
        set_ttl_settings=None,
        drop_ttl_settings=None,
        add_column_families=None,
        alter_column_families=None,
        alter_storage_settings=None,
        set_compaction_policy=None,
        alter_partitioning_settings=None,
        set_key_bloom_filter=None,
        set_read_replicas_settings=None,
    ):
        return self._driver.future(
            _session_impl.alter_table_request_factory(
                self._state,
                path,
                add_columns,
                drop_columns,
                alter_attributes,
                add_indexes,
                drop_indexes,
                set_ttl_settings,
                drop_ttl_settings,
                add_column_families,
                alter_column_families,
                alter_storage_settings,
                set_compaction_policy,
                alter_partitioning_settings,
                set_key_bloom_filter,
                set_read_replicas_settings,
            ),
            _apis.TableService.Stub,
            _apis.TableService.AlterTable,
            _session_impl.AlterTableOperation,
            settings,
            (self._driver,),
            self._state.endpoint,
        )

    def async_copy_table(self, source_path, destination_path, settings=None):
        return self.async_copy_tables(
            [(source_path, destination_path)], settings=settings
        )

    @_utilities.wrap_async_call_exceptions
    def async_copy_tables(self, source_destination_pairs, settings=None):
        return self._driver.future(
            _session_impl.copy_tables_request_factory(
                self._state, source_destination_pairs
            ),
            _apis.TableService.Stub,
            _apis.TableService.CopyTables,
            _session_impl.wrap_operation,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_rename_tables(self, rename_tables, settings=None):
        return self._driver.future(
            _session_impl.rename_tables_request_factory(self._state, rename_tables),
            _apis.TableService.Stub,
            _apis.TableService.RenameTables,
            _session_impl.wrap_operation,
            settings,
            (self._state,),
            self._state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_describe_table(self, path, settings=None):
        return self._driver.future(
            _session_impl.describe_table_request_factory(self._state, path, settings),
            _apis.TableService.Stub,
            _apis.TableService.DescribeTable,
            _session_impl.wrap_describe_table_response,
            settings,
            (self._state, TableSchemeEntry),
            self._state.endpoint,
        )


@six.add_metaclass(abc.ABCMeta)
class ITxContext:
    @abstractmethod
    def __init__(self, driver, session_state, session, tx_mode=None):
        """
        An object that provides a simple transaction context manager that allows statements execution
        in a transaction. You don't have to open transaction explicitly, because context manager encapsulates
        transaction control logic, and opens new transaction if:
         1) By explicit .begin();
         2) On execution of a first statement, which is strictly recommended method, because that avoids
         useless round trip

        This context manager is not thread-safe, so you should not manipulate on it concurrently.

        :param driver: A driver instance
        :param session_state: A state of session
        :param tx_mode: A transaction mode, which is a one from the following choices:
         1) SerializableReadWrite() which is default mode;
         2) OnlineReadOnly();
         3) StaleReadOnly().
        """
        pass

    @abstractmethod
    def __enter__(self):
        """
        Enters a context manager and returns a session

        :return: A session instance
        """
        pass

    @abstractmethod
    def __exit__(self, *args, **kwargs):
        """
        Closes a transaction context manager and rollbacks transaction if
        it is not rolled back explicitly
        """
        pass

    @property
    @abstractmethod
    def session_id(self):
        """
        A transaction's session id

        :return: A transaction's session id
        """
        pass

    @property
    @abstractmethod
    def tx_id(self):
        """
        Returns a id of open transaction or None otherwise

        :return: A id of open transaction or None otherwise
        """
        pass

    @abstractmethod
    def execute(self, query, parameters=None, commit_tx=False, settings=None):
        """
        Sends a query (yql text or an instance of DataQuery) to be executed with parameters.
        Execution with parameters supported only for DataQuery instances and is not supported yql text queries.

        :param query: A query, yql text or DataQuery instance.
        :param parameters: A dictionary with parameters values.
        :param commit_tx: A special flag that allows transaction commit
        :param settings: An additional request settings

        :return: A result sets or exception in case of execution errors
        """
        pass

    @abstractmethod
    def commit(self, settings=None):
        """
        Calls commit on a transaction if it is open otherwise is no-op. If transaction execution
        failed then this method raises PreconditionFailed.

        :param settings: A request settings

        :return: A committed transaction or exception if commit is failed
        """
        pass

    @abstractmethod
    def rollback(self, settings=None):
        """
        Calls rollback on a transaction if it is open otherwise is no-op. If transaction execution
        failed then this method raises PreconditionFailed.

        :param settings: A request settings

        :return: A rolled back transaction or exception if rollback is failed
        """
        pass

    @abstractmethod
    def begin(self, settings=None):
        """
        Explicitly begins a transaction

        :param settings: A request settings

        :return: An open transaction
        """
        pass


class BaseTxContext(ITxContext):
    __slots__ = ("_tx_state", "_session_state", "_driver", "session")

    def __init__(self, driver, session_state, session, tx_mode=None):
        """
        An object that provides a simple transaction context manager that allows statements execution
        in a transaction. You don't have to open transaction explicitly, because context manager encapsulates
        transaction control logic, and opens new transaction if:

        1) By explicit .begin() and .async_begin() methods;
        2) On execution of a first statement, which is strictly recommended method, because that avoids useless round trip

        This context manager is not thread-safe, so you should not manipulate on it concurrently.

        :param driver: A driver instance
        :param session_state: A state of session
        :param tx_mode: A transaction mode, which is a one from the following choices:
         1) SerializableReadWrite() which is default mode;
         2) OnlineReadOnly();
         3) StaleReadOnly().
        """
        self._driver = driver
        tx_mode = SerializableReadWrite() if tx_mode is None else tx_mode
        self._tx_state = _tx_ctx_impl.TxState(tx_mode)
        self._session_state = session_state
        self.session = session

    def __enter__(self):
        """
        Enters a context manager and returns a session

        :return: A session instance
        """
        return self

    def __exit__(self, *args, **kwargs):
        """
        Closes a transaction context manager and rollbacks transaction if
        it is not rolled back explicitly
        """
        if self._tx_state.tx_id is not None:
            # It's strictly recommended to close transactions directly
            # by using commit_tx=True flag while executing statement or by
            # .commit() or .rollback() methods, but here we trying to do best
            # effort to avoid useless open transactions
            logger.warning("Potentially leaked tx: %s", self._tx_state.tx_id)
            try:
                self.rollback()
            except issues.Error:
                logger.warning("Failed to rollback leaked tx: %s", self._tx_state.tx_id)

            self._tx_state.tx_id = None

    @property
    def session_id(self):
        """
        A transaction's session id

        :return: A transaction's session id
        """
        return self._session_state.session_id

    @property
    def tx_id(self):
        """
        Returns a id of open transaction or None otherwise

        :return: A id of open transaction or None otherwise
        """
        return self._tx_state.tx_id

    def execute(self, query, parameters=None, commit_tx=False, settings=None):
        """
        Sends a query (yql text or an instance of DataQuery) to be executed with parameters.
        Execution with parameters supported only for DataQuery instances and is not supported yql text queries.

        :param query: A query, yql text or DataQuery instance.
        :param parameters: A dictionary with parameters values.
        :param commit_tx: A special flag that allows transaction commit
        :param settings: An additional request settings

        :return: A result sets or exception in case of execution errors
        """
        return self._driver(
            _tx_ctx_impl.execute_request_factory(
                self._session_state,
                self._tx_state,
                query,
                parameters,
                commit_tx,
                settings,
            ),
            _apis.TableService.Stub,
            _apis.TableService.ExecuteDataQuery,
            _tx_ctx_impl.wrap_result_and_tx_id,
            settings,
            (self._session_state, self._tx_state, query),
            self._session_state.endpoint,
        )

    def commit(self, settings=None):
        """
        Calls commit on a transaction if it is open otherwise is no-op. If transaction execution
        failed then this method raises PreconditionFailed.

        :param settings: A request settings

        :return: A committed transaction or exception if commit is failed
        """
        if self._tx_state.tx_id is None and not self._tx_state.dead:
            return self
        return self._driver(
            _tx_ctx_impl.commit_request_factory(self._session_state, self._tx_state),
            _apis.TableService.Stub,
            _apis.TableService.CommitTransaction,
            _tx_ctx_impl.wrap_result_on_rollback_or_commit_tx,
            settings,
            (self._session_state, self._tx_state, self),
            self._session_state.endpoint,
        )

    def rollback(self, settings=None):
        """
        Calls rollback on a transaction if it is open otherwise is no-op. If transaction execution
        failed then this method raises PreconditionFailed.

        :param settings: A request settings

        :return: A rolled back transaction or exception if rollback is failed
        """
        if self._tx_state.tx_id is None and not self._tx_state.dead:
            return self
        return self._driver(
            _tx_ctx_impl.rollback_request_factory(self._session_state, self._tx_state),
            _apis.TableService.Stub,
            _apis.TableService.RollbackTransaction,
            _tx_ctx_impl.wrap_result_on_rollback_or_commit_tx,
            settings,
            (self._session_state, self._tx_state, self),
            self._session_state.endpoint,
        )

    def begin(self, settings=None):
        """
        Explicitly begins a transaction

        :param settings: A request settings

        :return: An open transaction
        """
        if self._tx_state.tx_id is not None:
            return self
        return self._driver(
            _tx_ctx_impl.begin_request_factory(self._session_state, self._tx_state),
            _apis.TableService.Stub,
            _apis.TableService.BeginTransaction,
            _tx_ctx_impl.wrap_tx_begin_response,
            settings,
            (self._session_state, self._tx_state, self),
            self._session_state.endpoint,
        )


class TxContext(BaseTxContext):
    @_utilities.wrap_async_call_exceptions
    def async_execute(self, query, parameters=None, commit_tx=False, settings=None):
        """
        Sends a query (yql text or an instance of DataQuery) to be executed with parameters.
        Execution with parameters supported only for DataQuery instances and not supported for YQL text.

        :param query: A query: YQL text or DataQuery instance. E
        :param parameters: A dictionary with parameters values.
        :param commit_tx: A special flag that allows transaction commit
        :param settings: A request settings (an instance of ExecDataQuerySettings)

        :return: A future of query execution
        """
        return self._driver.future(
            _tx_ctx_impl.execute_request_factory(
                self._session_state,
                self._tx_state,
                query,
                parameters,
                commit_tx,
                settings,
            ),
            _apis.TableService.Stub,
            _apis.TableService.ExecuteDataQuery,
            _tx_ctx_impl.wrap_result_and_tx_id,
            settings,
            (
                self._session_state,
                self._tx_state,
                query,
            ),
            self._session_state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_commit(self, settings=None):
        """
        Calls commit on a transaction if it is open otherwise is no-op. If transaction execution
        failed then this method raises PreconditionFailed.

        :param settings: A request settings (an instance of BaseRequestSettings)

        :return: A future of commit call
        """
        if self._tx_state.tx_id is None and not self._tx_state.dead:
            return _utilities.wrap_result_in_future(self)
        return self._driver.future(
            _tx_ctx_impl.commit_request_factory(self._session_state, self._tx_state),
            _apis.TableService.Stub,
            _apis.TableService.CommitTransaction,
            _tx_ctx_impl.wrap_result_on_rollback_or_commit_tx,
            settings,
            (self._session_state, self._tx_state, self),
            self._session_state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_rollback(self, settings=None):
        """
        Calls rollback on a transaction if it is open otherwise is no-op. If transaction execution
        failed then this method raises PreconditionFailed.

        :param settings: A request settings

        :return: A future of rollback call
        """
        if self._tx_state.tx_id is None and not self._tx_state.dead:
            return _utilities.wrap_result_in_future(self)
        return self._driver.future(
            _tx_ctx_impl.rollback_request_factory(self._session_state, self._tx_state),
            _apis.TableService.Stub,
            _apis.TableService.RollbackTransaction,
            _tx_ctx_impl.wrap_result_on_rollback_or_commit_tx,
            settings,
            (self._session_state, self._tx_state, self),
            self._session_state.endpoint,
        )

    @_utilities.wrap_async_call_exceptions
    def async_begin(self, settings=None):
        """
        Explicitly begins a transaction

        :param settings: A request settings

        :return: A future of begin call
        """
        if self._tx_state.tx_id is not None:
            return _utilities.wrap_result_in_future(self)
        return self._driver.future(
            _tx_ctx_impl.begin_request_factory(self._session_state, self._tx_state),
            _apis.TableService.Stub,
            _apis.TableService.BeginTransaction,
            _tx_ctx_impl.wrap_tx_begin_response,
            settings,
            (self._session_state, self._tx_state, self),
            self._session_state.endpoint,
        )


class SessionPool(object):
    def __init__(
        self,
        driver,
        size=100,
        workers_threads_count=4,
        initializer=None,
        min_pool_size=0,
    ):
        """
        An object that encapsulates session creation, deletion and etc. and maintains
        a pool of active sessions of specified size

        :param driver: A Driver instance
        :param size: A maximum number of sessions to maintain in the pool
        """
        self._logger = logger.getChild(self.__class__.__name__)
        self._pool_impl = _sp_impl.SessionPoolImpl(
            self._logger,
            driver,
            size,
            workers_threads_count,
            initializer,
            min_pool_size,
        )
        if hasattr(driver, "_driver_config"):
            self.tracer = driver._driver_config.tracer
        else:
            self.tracer = ydb.Tracer(None)

    def retry_operation_sync(self, callee, retry_settings=None, *args, **kwargs):

        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        def wrapped_callee():
            with self.checkout(
                timeout=retry_settings.get_session_client_timeout
            ) as session:
                return callee(session, *args, **kwargs)

        return retry_operation_sync(wrapped_callee, retry_settings)

    @property
    def active_size(self):
        return self._pool_impl.active_size

    @property
    def free_size(self):
        return self._pool_impl.free_size

    @property
    def busy_size(self):
        return self._pool_impl.busy_size

    @property
    def max_size(self):
        return self._pool_impl.max_size

    @property
    def waiters_count(self):
        return self._pool_impl.waiters_count

    @tracing.with_trace()
    def subscribe(self):
        return self._pool_impl.subscribe()

    @tracing.with_trace()
    def unsubscribe(self, waiter):
        return self._pool_impl.unsubscribe(waiter)

    @tracing.with_trace()
    def acquire(self, blocking=True, timeout=None):
        return self._pool_impl.acquire(blocking, timeout)

    @tracing.with_trace()
    def release(self, session):
        return self._pool_impl.put(session)

    def async_checkout(self):
        """
        Returns a context manager that asynchronously checkouts a session from the pool.

        """
        return AsyncSessionCheckout(self)

    def checkout(self, blocking=True, timeout=None):
        return SessionCheckout(self, blocking, timeout)

    def stop(self, timeout=None):
        self._pool_impl.stop(timeout)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class AsyncSessionCheckout(object):
    __slots__ = ("subscription", "pool")

    def __init__(self, pool):
        """
        A context manager that asynchronously checkouts a session for the specified pool
        and returns it on manager exit.

        :param pool: A SessionPool instance.
        """
        self.pool = pool
        self.subscription = None

    def __enter__(self):
        self.subscription = self.pool.subscribe()
        return self.subscription

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pool.unsubscribe(self.subscription)


class SessionCheckout(object):
    __slots__ = ("_acquired", "_pool", "_blocking", "_timeout")

    def __init__(self, pool, blocking, timeout):
        """
        A context manager that checkouts a session from the specified pool and
        returns it on manager exit.

        :param pool: A SessionPool instance
        :param blocking: A flag that specifies that session acquire method should blocks
        :param timeout: A timeout in seconds for session acquire
        """
        self._pool = pool
        self._acquired = None
        self._blocking = blocking
        self._timeout = timeout

    def __enter__(self):
        self._acquired = self._pool.acquire(self._blocking, self._timeout)
        return self._acquired

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._acquired is not None:
            self._pool.release(self._acquired)



class AsyncResponseIterator(object):
    def __init__(self, it, wrapper):
        self.it = it.__aiter__()
        self.wrapper = wrapper

    def cancel(self):
        self.it.cancel()
        return self

    def __aiter__(self):
        return self

    async def _next(self):
        return self.wrapper(await self.it.__anext__())

    async def next(self):
        return await self._next()

    async def __anext__(self):
        return await self._next()



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_topic_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_topic_pb2 as protos_dot_ydb__topic__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_topic_v1.proto',
  package='Ydb.Topic.V1',
  syntax='proto3',
  serialized_options=b'\n\021tech.ydb.topic.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Topic_V1\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12ydb_topic_v1.proto\x12\x0cYdb.Topic.V1\x1a\x16protos/ydb_topic.proto2\xeb\x04\n\x0cTopicService\x12\x65\n\x0bStreamWrite\x12(.Ydb.Topic.StreamWriteMessage.FromClient\x1a(.Ydb.Topic.StreamWriteMessage.FromServer(\x01\x30\x01\x12\x62\n\nStreamRead\x12\'.Ydb.Topic.StreamReadMessage.FromClient\x1a\'.Ydb.Topic.StreamReadMessage.FromServer(\x01\x30\x01\x12L\n\x0b\x43reateTopic\x12\x1d.Ydb.Topic.CreateTopicRequest\x1a\x1e.Ydb.Topic.CreateTopicResponse\x12R\n\rDescribeTopic\x12\x1f.Ydb.Topic.DescribeTopicRequest\x1a .Ydb.Topic.DescribeTopicResponse\x12[\n\x10\x44\x65scribeConsumer\x12\".Ydb.Topic.DescribeConsumerRequest\x1a#.Ydb.Topic.DescribeConsumerResponse\x12I\n\nAlterTopic\x12\x1c.Ydb.Topic.AlterTopicRequest\x1a\x1d.Ydb.Topic.AlterTopicResponse\x12\x46\n\tDropTopic\x12\x1b.Ydb.Topic.DropTopicRequest\x1a\x1c.Ydb.Topic.DropTopicResponseBL\n\x11tech.ydb.topic.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Topic_V1\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[protos_dot_ydb__topic__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_TOPICSERVICE = _descriptor.ServiceDescriptor(
  name='TopicService',
  full_name='Ydb.Topic.V1.TopicService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=61,
  serialized_end=680,
  methods=[
  _descriptor.MethodDescriptor(
    name='StreamWrite',
    full_name='Ydb.Topic.V1.TopicService.StreamWrite',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._STREAMWRITEMESSAGE_FROMCLIENT,
    output_type=protos_dot_ydb__topic__pb2._STREAMWRITEMESSAGE_FROMSERVER,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='StreamRead',
    full_name='Ydb.Topic.V1.TopicService.StreamRead',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._STREAMREADMESSAGE_FROMCLIENT,
    output_type=protos_dot_ydb__topic__pb2._STREAMREADMESSAGE_FROMSERVER,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateTopic',
    full_name='Ydb.Topic.V1.TopicService.CreateTopic',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._CREATETOPICREQUEST,
    output_type=protos_dot_ydb__topic__pb2._CREATETOPICRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DescribeTopic',
    full_name='Ydb.Topic.V1.TopicService.DescribeTopic',
    index=3,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._DESCRIBETOPICREQUEST,
    output_type=protos_dot_ydb__topic__pb2._DESCRIBETOPICRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DescribeConsumer',
    full_name='Ydb.Topic.V1.TopicService.DescribeConsumer',
    index=4,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._DESCRIBECONSUMERREQUEST,
    output_type=protos_dot_ydb__topic__pb2._DESCRIBECONSUMERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AlterTopic',
    full_name='Ydb.Topic.V1.TopicService.AlterTopic',
    index=5,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._ALTERTOPICREQUEST,
    output_type=protos_dot_ydb__topic__pb2._ALTERTOPICRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DropTopic',
    full_name='Ydb.Topic.V1.TopicService.DropTopic',
    index=6,
    containing_service=None,
    input_type=protos_dot_ydb__topic__pb2._DROPTOPICREQUEST,
    output_type=protos_dot_ydb__topic__pb2._DROPTOPICRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TOPICSERVICE)

DESCRIPTOR.services_by_name['TopicService'] = _TOPICSERVICE

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
import decimal
from google.protobuf import struct_pb2
import six

from . import issues, types, _apis


_SHIFT_BIT_COUNT = 64
_SHIFT = 2**64
_SIGN_BIT = 2**63
_DecimalNanRepr = 10**35 + 1
_DecimalInfRepr = 10**35
_DecimalSignedInfRepr = -(10**35)
_primitive_type_by_id = {}


def _initialize():
    for pt in types.PrimitiveType:
        _primitive_type_by_id[pt._idn_] = pt


_initialize()


class _DotDict(dict):
    def __init__(self, *args, **kwargs):
        super(_DotDict, self).__init__(*args, **kwargs)

    def __getattr__(self, item):
        return self[item]


def _is_decimal_signed(hi_value):
    return (hi_value & _SIGN_BIT) == _SIGN_BIT


def _pb_to_decimal(type_pb, value_pb, table_client_settings):
    hi = (
        (value_pb.high_128 - (1 << _SHIFT_BIT_COUNT))
        if _is_decimal_signed(value_pb.high_128)
        else value_pb.high_128
    )
    int128_value = value_pb.low_128 + (hi << _SHIFT_BIT_COUNT)
    if int128_value == _DecimalNanRepr:
        return decimal.Decimal("Nan")
    elif int128_value == _DecimalInfRepr:
        return decimal.Decimal("Inf")
    elif int128_value == _DecimalSignedInfRepr:
        return decimal.Decimal("-Inf")
    return decimal.Decimal(int128_value) / decimal.Decimal(
        10**type_pb.decimal_type.scale
    )


def _pb_to_primitive(type_pb, value_pb, table_client_settings):
    return _primitive_type_by_id.get(type_pb.type_id).get_value(
        value_pb, table_client_settings
    )


def _pb_to_optional(type_pb, value_pb, table_client_settings):
    if value_pb.WhichOneof("value") == "null_flag_value":
        return None
    if value_pb.WhichOneof("value") == "nested_value":
        return _to_native_value(
            type_pb.optional_type.item, value_pb.nested_value, table_client_settings
        )
    return _to_native_value(type_pb.optional_type.item, value_pb, table_client_settings)


def _pb_to_list(type_pb, value_pb, table_client_settings):
    return [
        _to_native_value(
            type_pb.list_type.item, value_proto_item, table_client_settings
        )
        for value_proto_item in value_pb.items
    ]


def _pb_to_tuple(type_pb, value_pb, table_client_settings):
    return tuple(
        _to_native_value(item_type, item_value, table_client_settings)
        for item_type, item_value in six.moves.zip(
            type_pb.tuple_type.elements, value_pb.items
        )
    )


def _pb_to_dict(type_pb, value_pb, table_client_settings):
    result = {}
    for kv_pair in value_pb.pairs:
        key = _to_native_value(
            type_pb.dict_type.key, kv_pair.key, table_client_settings
        )
        payload = _to_native_value(
            type_pb.dict_type.payload, kv_pair.payload, table_client_settings
        )
        result[key] = payload
    return result


class _Struct(_DotDict):
    pass


def _pb_to_struct(type_pb, value_pb, table_client_settings):
    result = _Struct()
    for member, item in six.moves.zip(type_pb.struct_type.members, value_pb.items):
        result[member.name] = _to_native_value(member.type, item, table_client_settings)
    return result


def _pb_to_void(type_pb, value_pb, table_client_settings):
    return None


_to_native_map = {
    "type_id": _pb_to_primitive,
    "decimal_type": _pb_to_decimal,
    "optional_type": _pb_to_optional,
    "list_type": _pb_to_list,
    "tuple_type": _pb_to_tuple,
    "dict_type": _pb_to_dict,
    "struct_type": _pb_to_struct,
    "void_type": _pb_to_void,
}


def _to_native_value(type_pb, value_pb, table_client_settings=None):
    return _to_native_map.get(type_pb.WhichOneof("type"))(
        type_pb, value_pb, table_client_settings
    )


def _decimal_to_int128(value_type, value):
    if value.is_nan():
        return _DecimalNanRepr
    elif value.is_infinite():
        if value.is_signed():
            return _DecimalSignedInfRepr
        return _DecimalInfRepr

    sign, digits, exponent = value.as_tuple()
    int128_value = 0
    digits_count = 0
    for digit in digits:
        int128_value *= 10
        int128_value += digit
        digits_count += 1

    if value_type.decimal_type.scale + exponent < 0:
        raise issues.GenericError("Couldn't parse decimal value, exponent is too large")

    for _ in range(value_type.decimal_type.scale + exponent):
        int128_value *= 10
        digits_count += 1

    if digits_count > value_type.decimal_type.precision + value_type.decimal_type.scale:
        raise issues.GenericError("Couldn't parse decimal value, digits count > 35")

    if sign:
        int128_value *= -1

    return int128_value


def _decimal_to_pb(value_type, value):
    value_pb = _apis.ydb_value.Value()
    int128_value = _decimal_to_int128(value_type, value)
    if int128_value < 0:
        value_pb.high_128 = (int128_value >> _SHIFT_BIT_COUNT) + (1 << _SHIFT_BIT_COUNT)
        int128_value -= (int128_value >> _SHIFT_BIT_COUNT) << _SHIFT_BIT_COUNT
    else:
        value_pb.high_128 = int128_value >> _SHIFT_BIT_COUNT
        int128_value -= value_pb.high_128 << _SHIFT_BIT_COUNT
    value_pb.low_128 = int128_value
    return value_pb


def _primitive_to_pb(type_pb, value):
    value_pb = _apis.ydb_value.Value()
    data_type = _primitive_type_by_id.get(type_pb.type_id)
    data_type.set_value(value_pb, value)
    return value_pb


def _optional_to_pb(type_pb, value):
    if value is None:
        return _apis.ydb_value.Value(null_flag_value=struct_pb2.NULL_VALUE)
    return _from_native_value(type_pb.optional_type.item, value)


def _list_to_pb(type_pb, value):
    value_pb = _apis.ydb_value.Value()
    for element in value:
        value_item_proto = value_pb.items.add()
        value_item_proto.MergeFrom(_from_native_value(type_pb.list_type.item, element))
    return value_pb


def _tuple_to_pb(type_pb, value):
    value_pb = _apis.ydb_value.Value()
    for element_type, element_value in six.moves.zip(
        type_pb.tuple_type.elements, value
    ):
        value_item_proto = value_pb.items.add()
        value_item_proto.MergeFrom(_from_native_value(element_type, element_value))
    return value_pb


def _dict_to_pb(type_pb, value):
    value_pb = _apis.ydb_value.Value()
    for key, payload in value.items():
        kv_pair = value_pb.pairs.add()
        kv_pair.key.MergeFrom(_from_native_value(type_pb.dict_type.key, key))
        kv_pair.payload.MergeFrom(
            _from_native_value(type_pb.dict_type.payload, payload)
        )
    return value_pb


def _struct_to_pb(type_pb, value):
    value_pb = _apis.ydb_value.Value()
    for member in type_pb.struct_type.members:
        value_item_proto = value_pb.items.add()
        value_item = (
            value[member.name]
            if isinstance(value, dict)
            else getattr(value, member.name)
        )
        value_item_proto.MergeFrom(_from_native_value(member.type, value_item))
    return value_pb


_from_native_map = {
    "type_id": _primitive_to_pb,
    "decimal_type": _decimal_to_pb,
    "optional_type": _optional_to_pb,
    "list_type": _list_to_pb,
    "tuple_type": _tuple_to_pb,
    "dict_type": _dict_to_pb,
    "struct_type": _struct_to_pb,
}


def _decimal_type_to_native(type_pb):
    return types.DecimalType(type_pb.decimal_type.precision, type_pb.decimal_type.scale)


def _optional_type_to_native(type_pb):
    return types.OptionalType(type_to_native(type_pb.optional_type.item))


def _primitive_type_to_native(type_pb):
    return _primitive_type_by_id.get(type_pb.type_id)


def _null_type_factory(type_pb):
    return types.NullType()


_type_to_native_map = {
    "optional_type": _optional_type_to_native,
    "type_id": _primitive_type_to_native,
    "decimal_type": _decimal_type_to_native,
    "null_type": _null_type_factory,
}


def type_to_native(type_pb):
    return _type_to_native_map.get(type_pb.WhichOneof("type"))(type_pb)


def _from_native_value(type_pb, value):
    return _from_native_map.get(type_pb.WhichOneof("type"))(type_pb, value)


def to_typed_value_from_native(type_pb, value):
    typed_value = _apis.ydb_value.TypedValue()
    typed_value.type.MergeFrom(type_pb)
    typed_value.value.MergeFrom(from_native_value(type_pb, value))
    return typed_value


def parameters_to_pb(parameters_types, parameters_values):
    if parameters_values is None or not parameters_values:
        return {}

    param_values_pb = {}
    for name, type_pb in six.iteritems(parameters_types):
        result = _apis.ydb_value.TypedValue()
        ttype = type_pb
        if isinstance(type_pb, types.AbstractTypeBuilder):
            ttype = type_pb.proto
        elif isinstance(type_pb, types.PrimitiveType):
            ttype = type_pb.proto
        result.type.MergeFrom(ttype)
        result.value.MergeFrom(_from_native_value(ttype, parameters_values[name]))
        param_values_pb[name] = result
    return param_values_pb


def _unwrap_optionality(column):
    c_type = column.type
    current_type = c_type.WhichOneof("type")
    while current_type == "optional_type":
        c_type = c_type.optional_type.item
        current_type = c_type.WhichOneof("type")
    return _to_native_map.get(current_type), c_type


class _ResultSet(object):
    __slots__ = ("columns", "rows", "truncated")

    def __init__(self, columns, rows, truncated):
        self.columns = columns
        self.rows = rows
        self.truncated = truncated

    @classmethod
    def from_message(cls, message, table_client_settings=None):
        rows = []
        # prepare columnn parsers before actuall parsing
        column_parsers = []
        if len(message.rows) > 0:
            for column in message.columns:
                column_parsers.append(_unwrap_optionality(column))

        for row_proto in message.rows:
            row = _Row(message.columns)
            for column, value, column_info in six.moves.zip(
                message.columns, row_proto.items, column_parsers
            ):
                v_type = value.WhichOneof("value")
                if v_type == "null_flag_value":
                    row[column.name] = None
                    continue

                while v_type == "nested_value":
                    value = value.nested_value
                    v_type = value.WhichOneof("value")

                column_parser, unwrapped_type = column_info
                row[column.name] = column_parser(
                    unwrapped_type, value, table_client_settings
                )
            rows.append(row)
        return cls(message.columns, rows, message.truncated)

    @classmethod
    def lazy_from_message(cls, message, table_client_settings=None):
        rows = _LazyRows(message.rows, table_client_settings, message.columns)
        return cls(message.columns, rows, message.truncated)


ResultSet = _ResultSet


class _Row(_DotDict):
    def __init__(self, columns):
        super(_Row, self).__init__()
        self._columns = columns

    def __getitem__(self, key):
        if isinstance(key, int):
            return self[self._columns[key].name]
        elif isinstance(key, slice):
            return tuple(map(lambda x: self[x.name], self._columns[key]))
        else:
            return super(_Row, self).__getitem__(key)


class _LazyRowItem:

    __slots__ = ["_item", "_type", "_table_client_settings", "_processed", "_parser"]

    def __init__(self, proto_item, proto_type, table_client_settings, parser):
        self._item = proto_item
        self._type = proto_type
        self._table_client_settings = table_client_settings
        self._processed = False
        self._parser = parser

    def get(self):
        if not self._processed:

            self._item = self._parser(
                self._type, self._item, self._table_client_settings
            )
            self._processed = True
        return self._item


class _LazyRow(_DotDict):
    def __init__(self, columns, proto_row, table_client_settings, parsers):
        super(_LazyRow, self).__init__()
        self._columns = columns
        self._table_client_settings = table_client_settings
        for i, (column, row_item) in enumerate(
            six.moves.zip(self._columns, proto_row.items)
        ):
            super(_LazyRow, self).__setitem__(
                column.name,
                _LazyRowItem(row_item, column.type, table_client_settings, parsers[i]),
            )

    def __setitem__(self, key, value):
        raise NotImplementedError("Cannot insert values into lazy row")

    def __getitem__(self, key):
        if isinstance(key, int):
            return self[self._columns[key].name]
        elif isinstance(key, slice):
            return tuple(map(lambda x: self[x.name], self._columns[key]))
        else:
            return super(_LazyRow, self).__getitem__(key).get()

    def __iter__(self):
        return super(_LazyRow, self).__iter__()

    def __next__(self):
        return super(_LazyRow, self).__next__().get()

    def next(self):
        return self.__next__()


def from_native_value(type_pb, value):
    return _from_native_value(type_pb, value)


def to_native_value(typed_value):
    return _to_native_value(typed_value.type, typed_value.value)


class _LazyRows:
    def __init__(self, rows, table_client_settings, columns):
        self._rows = rows
        self._parsers = [_LazyParser(columns, i) for i in range(len(columns))]
        self._table_client_settings = table_client_settings
        self._columns = columns

    def __len__(self):
        return len(self._rows)

    def fetchone(self):
        return _LazyRow(
            self._columns, self._rows[0], self._table_client_settings, self._parsers
        )

    def fetchmany(self, number):
        for index in range(min(len(self), number)):
            yield _LazyRow(
                self._columns,
                self._rows[index],
                self._table_client_settings,
                self._parsers,
            )

    def __iter__(self):
        for row in self.fetchmany(len(self)):
            yield row

    def fetchall(self):
        for row in self:
            yield row


class _LazyParser:
    __slots__ = ["_columns", "_column_index", "_prepared"]

    def __init__(self, columns, column_index):
        self._columns = columns
        self._column_index = column_index
        self._prepared = None

    def __call__(self, *args, **kwargs):
        if self._prepared is None:
            self._prepared = _to_native_map.get(
                self._columns[self._column_index].type.WhichOneof("type")
            )
        return self._prepared(*args, **kwargs)


class ResultSets(list):
    def __init__(self, result_sets_pb, table_client_settings=None):
        make_lazy = (
            False
            if table_client_settings is None
            else table_client_settings._make_result_sets_lazy
        )
        result_sets = []
        initializer = (
            _ResultSet.from_message if not make_lazy else _ResultSet.lazy_from_message
        )
        for result_set in result_sets_pb:
            result_sets.append(initializer(result_set, table_client_settings))
        super(ResultSets, self).__init__(result_sets)



from enum import IntEnum
import functools


class TraceLevel(IntEnum):
    DEBUG = 0
    INFO = 1
    ERROR = 2
    NONE = 3


class _TracingCtx:
    def __init__(self, tracer, span_name):
        self._enabled = tracer._open_tracer is not None
        self._scope = None
        self._tracer = tracer
        self._span_name = span_name

    def __enter__(self):
        """
        Creates new span
        :return: self
        """
        if not self._enabled:
            return self
        self._scope = self._tracer._open_tracer.start_active_span(self._span_name)
        self._scope.span.set_baggage_item("ctx", self)
        self.trace(self._tracer._pre_tags)
        return self

    @property
    def enabled(self):
        """
        :return: Is tracing enabled
        """
        return self._enabled

    def trace(self, tags, trace_level=TraceLevel.INFO):
        """
        Add tags to current span

        :param ydb.TraceLevel trace_level: level of tracing
        :param dict tags: Dict of tags
        """
        if self._tracer._verbose_level < trace_level:
            return
        if not self.enabled or self._scope is None:
            return
        for key, value in tags.items():
            self._scope.span.set_tag(key, value)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.enabled:
            return
        if exc_val:
            self.trace(self._tracer._post_tags_err, trace_level=TraceLevel.ERROR)
            self._tracer._on_err(self, exc_type, exc_val, exc_tb)
        else:
            self.trace(self._tracer._post_tags_ok)
        self._scope.close()
        self._scope = None


def with_trace(span_name=None):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            name = (
                span_name
                if span_name is not None
                else self.__class__.__name__ + "." + f.__name__
            )
            with self.tracer.trace(name):
                return f(self, *args, **kwargs)

        return wrapper

    return decorator


def trace(tracer, tags, trace_level=TraceLevel.INFO):
    if tracer.enabled:
        scope = tracer._open_tracer.scope_manager.active
        if not scope:
            return False

        ctx = scope.span.get_baggage_item("ctx")
        if ctx is None:
            return False

        return ctx.trace(tags, trace_level)


class Tracer:
    def __init__(self, tracer):
        """
        Init an tracer to trace requests

        :param opentracing.Tracer tracer: opentracing.Tracer implementation. If None - tracing not enabled
        """
        self._open_tracer = tracer
        self._pre_tags = {}
        self._post_tags_ok = {}
        self._post_tags_err = {}
        self._on_err = lambda *args, **kwargs: None
        self._verbose_level = TraceLevel.NONE

    @property
    def enabled(self):
        return self._open_tracer is not None

    def trace(self, span_name):
        """
        Create tracing context

        :param str span_name:

        :return: A tracing context
        :rtype: _TracingCtx
        """
        return _TracingCtx(self, span_name)

    def with_pre_tags(self, tags):
        """
        Add `tags` to every span immediately after creation

        :param dict tags: tags dict

        :return: self
        """
        self._pre_tags = tags
        return self

    def with_post_tags(self, ok_tags, err_tags):
        """
        Add some tags before span close

        :param ok_tags: Add this tags if no error raised
        :param err_tags: Add this tags if there is an exception

        :return: self
        """
        self._post_tags_ok = ok_tags
        self._post_tags_err = err_tags
        return self

    def with_on_error_callback(self, callee):
        """
        Add an callback, that will be called if there is an exception in span

        :param callable[_TracingCtx, exc_type, exc_val, exc_tb] callee:

        :return: self
        """
        self._on_err = callee
        return self

    def with_verbose_level(self, level):
        self._verbose_level = level
        return self

    @classmethod
    def default(cls, tracer):
        """
        Create default tracer

        :param tracer:

        :return: new tracer
        """
        return (
            cls(tracer)
            .with_post_tags({"ok": True}, {"ok": False})
            .with_pre_tags({"started": True})
            .with_on_error_callback(_default_on_error_callback)
            .with_verbose_level(TraceLevel.INFO)
        )


def _default_on_error_callback(ctx, exc_type, exc_val, exc_tb):
    ctx.trace(
        {
            "error.type": exc_type.__name__,
            "error.value": exc_val,
            "error.traceback": exc_tb,
        },
        trace_level=TraceLevel.ERROR,
    )



import sys

import google.protobuf
from packaging.version import Version

# generated files are incompatible between 3 and 4 protobuf versions
# import right generated version for current protobuf lib
# sdk code must always import from ydb._grpc.common
protobuf_version = Version(google.protobuf.__version__)

if protobuf_version < Version("4.0"):
    from ydb._grpc.v3 import * # noqa
    from ydb._grpc.v3 import protos # noqa
    sys.modules["ydb._grpc.common"] = sys.modules["ydb._grpc.v3"]
    sys.modules["ydb._grpc.common.protos"] = sys.modules["ydb._grpc.v3.protos"]
else:
    from ydb._grpc.v4 import * # noqa
    from ydb._grpc.v4 import protos # noqa
    sys.modules["ydb._grpc.common"] = sys.modules["ydb._grpc.v4"]
    sys.modules["ydb._grpc.common.protos"] = sys.modules["ydb._grpc.v4.protos"]



import os

from . import pool, scheme, table
import ydb
from ydb.driver import get_config


def default_credentials(credentials=None):
    if credentials is not None:
        return credentials

    service_account_key_file = os.getenv("YDB_SERVICE_ACCOUNT_KEY_FILE_CREDENTIALS")
    if service_account_key_file is not None:
        from .iam import ServiceAccountCredentials

        return ServiceAccountCredentials.from_file(service_account_key_file)

    anonymous_credetials = os.getenv("YDB_ANONYMOUS_CREDENTIALS", "0") == "1"
    if anonymous_credetials:
        return ydb.credentials.AnonymousCredentials()

    metadata_credentials = os.getenv("YDB_METADATA_CREDENTIALS", "0") == "1"
    if metadata_credentials:
        from .iam import MetadataUrlCredentials

        return MetadataUrlCredentials()

    access_token = os.getenv("YDB_ACCESS_TOKEN_CREDENTIALS")
    if access_token is not None:
        return ydb.credentials.AccessTokenCredentials(access_token)

    # (legacy instantiation)
    creds = ydb.auth_helpers.construct_credentials_from_environ()
    if creds is not None:
        return creds

    from .iam import MetadataUrlCredentials

    return MetadataUrlCredentials()


class DriverConfig(ydb.DriverConfig):
    @classmethod
    def default_from_endpoint_and_database(
        cls, endpoint, database, root_certificates=None, credentials=None, **kwargs
    ):
        return cls(
            endpoint,
            database,
            credentials=default_credentials(credentials),
            root_certificates=root_certificates,
            **kwargs
        )

    @classmethod
    def default_from_connection_string(
        cls, connection_string, root_certificates=None, credentials=None, **kwargs
    ):
        endpoint, database = ydb.parse_connection_string(connection_string)
        return cls(
            endpoint,
            database,
            credentials=default_credentials(credentials),
            root_certificates=root_certificates,
            **kwargs
        )


class Driver(pool.ConnectionPool):
    def __init__(
        self,
        driver_config=None,
        connection_string=None,
        endpoint=None,
        database=None,
        root_certificates=None,
        credentials=None,
        **kwargs
    ):
        config = get_config(
            driver_config,
            connection_string,
            endpoint,
            database,
            root_certificates,
            credentials,
            config_class=DriverConfig,
        )

        super(Driver, self).__init__(config)

        self.scheme_client = scheme.SchemeClient(self)
        self.table_client = table.TableClient(self, config.table_client_settings)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_import.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v4.protos.annotations import validation_pb2 as protos_dot_annotations_dot_validation__pb2
from ydb._grpc.v4.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17protos/ydb_import.proto\x12\nYdb.Import\x1a#protos/annotations/validation.proto\x1a\x1aprotos/ydb_operation.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xcd\x01\n\x0eImportProgress\"\xba\x01\n\x08Progress\x12\x18\n\x14PROGRESS_UNSPECIFIED\x10\x00\x12\x16\n\x12PROGRESS_PREPARING\x10\x01\x12\x1a\n\x16PROGRESS_TRANSFER_DATA\x10\x02\x12\x1a\n\x16PROGRESS_BUILD_INDEXES\x10\x03\x12\x11\n\rPROGRESS_DONE\x10\x04\x12\x19\n\x15PROGRESS_CANCELLATION\x10\x05\x12\x16\n\x12PROGRESS_CANCELLED\x10\x06\"\xa0\x01\n\x12ImportItemProgress\x12\x13\n\x0bparts_total\x18\x01 \x01(\r\x12\x17\n\x0fparts_completed\x18\x02 \x01(\r\x12.\n\nstart_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x9d\x03\n\x14ImportFromS3Settings\x12\x16\n\x08\x65ndpoint\x18\x01 \x01(\tB\x04\x90\xe6*\x01\x12\x37\n\x06scheme\x18\x02 \x01(\x0e\x32\'.Ydb.Import.ImportFromS3Settings.Scheme\x12\x14\n\x06\x62ucket\x18\x03 \x01(\tB\x04\x90\xe6*\x01\x12\x18\n\naccess_key\x18\x04 \x01(\tB\x04\x90\xe6*\x01\x12\x18\n\nsecret_key\x18\x05 \x01(\tB\x04\x90\xe6*\x01\x12<\n\x05items\x18\x06 \x03(\x0b\x32%.Ydb.Import.ImportFromS3Settings.ItemB\x06\x9a\xe6*\x02(\x01\x12\x1c\n\x0b\x64\x65scription\x18\x07 \x01(\tB\x07\xa2\xe6*\x03\x18\x80\x01\x12\x19\n\x11number_of_retries\x18\x08 \x01(\r\x1a\x43\n\x04Item\x12\x1b\n\rsource_prefix\x18\x01 \x01(\tB\x04\x90\xe6*\x01\x12\x1e\n\x10\x64\x65stination_path\x18\x02 \x01(\tB\x04\x90\xe6*\x01\".\n\x06Scheme\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04HTTP\x10\x01\x12\t\n\x05HTTPS\x10\x02\"\x14\n\x12ImportFromS3Result\"\xb9\x01\n\x14ImportFromS3Metadata\x12\x32\n\x08settings\x18\x01 \x01(\x0b\x32 .Ydb.Import.ImportFromS3Settings\x12\x35\n\x08progress\x18\x02 \x01(\x0e\x32#.Ydb.Import.ImportProgress.Progress\x12\x36\n\x0eitems_progress\x18\x03 \x03(\x0b\x32\x1e.Ydb.Import.ImportItemProgress\"\x8a\x01\n\x13ImportFromS3Request\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x38\n\x08settings\x18\x02 \x01(\x0b\x32 .Ydb.Import.ImportFromS3SettingsB\x04\x90\xe6*\x01\"D\n\x14ImportFromS3Response\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\" \n\rYdbDumpFormat\x12\x0f\n\x07\x63olumns\x18\x01 \x03(\t\"\x12\n\x10ImportDataResult\"\xae\x01\n\x11ImportDataRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x17\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x42\t\xa2\xe6*\x05\x18\x80\x80\x80\x04\x12-\n\x08ydb_dump\x18\x04 \x01(\x0b\x32\x19.Ydb.Import.YdbDumpFormatH\x00\x42\x08\n\x06\x66ormat\"B\n\x12ImportDataResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.OperationBP\n\x10tech.ydb.import_Z9github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Import\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_import_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\020tech.ydb.import_Z9github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Import\370\001\001'
  _IMPORTFROMS3SETTINGS_ITEM.fields_by_name['source_prefix']._options = None
  _IMPORTFROMS3SETTINGS_ITEM.fields_by_name['source_prefix']._serialized_options = b'\220\346*\001'
  _IMPORTFROMS3SETTINGS_ITEM.fields_by_name['destination_path']._options = None
  _IMPORTFROMS3SETTINGS_ITEM.fields_by_name['destination_path']._serialized_options = b'\220\346*\001'
  _IMPORTFROMS3SETTINGS.fields_by_name['endpoint']._options = None
  _IMPORTFROMS3SETTINGS.fields_by_name['endpoint']._serialized_options = b'\220\346*\001'
  _IMPORTFROMS3SETTINGS.fields_by_name['bucket']._options = None
  _IMPORTFROMS3SETTINGS.fields_by_name['bucket']._serialized_options = b'\220\346*\001'
  _IMPORTFROMS3SETTINGS.fields_by_name['access_key']._options = None
  _IMPORTFROMS3SETTINGS.fields_by_name['access_key']._serialized_options = b'\220\346*\001'
  _IMPORTFROMS3SETTINGS.fields_by_name['secret_key']._options = None
  _IMPORTFROMS3SETTINGS.fields_by_name['secret_key']._serialized_options = b'\220\346*\001'
  _IMPORTFROMS3SETTINGS.fields_by_name['items']._options = None
  _IMPORTFROMS3SETTINGS.fields_by_name['items']._serialized_options = b'\232\346*\002(\001'
  _IMPORTFROMS3SETTINGS.fields_by_name['description']._options = None
  _IMPORTFROMS3SETTINGS.fields_by_name['description']._serialized_options = b'\242\346*\003\030\200\001'
  _IMPORTFROMS3REQUEST.fields_by_name['settings']._options = None
  _IMPORTFROMS3REQUEST.fields_by_name['settings']._serialized_options = b'\220\346*\001'
  _IMPORTDATAREQUEST.fields_by_name['data']._options = None
  _IMPORTDATAREQUEST.fields_by_name['data']._serialized_options = b'\242\346*\005\030\200\200\200\004'
  _IMPORTPROGRESS._serialized_start=138
  _IMPORTPROGRESS._serialized_end=343
  _IMPORTPROGRESS_PROGRESS._serialized_start=157
  _IMPORTPROGRESS_PROGRESS._serialized_end=343
  _IMPORTITEMPROGRESS._serialized_start=346
  _IMPORTITEMPROGRESS._serialized_end=506
  _IMPORTFROMS3SETTINGS._serialized_start=509
  _IMPORTFROMS3SETTINGS._serialized_end=922
  _IMPORTFROMS3SETTINGS_ITEM._serialized_start=807
  _IMPORTFROMS3SETTINGS_ITEM._serialized_end=874
  _IMPORTFROMS3SETTINGS_SCHEME._serialized_start=876
  _IMPORTFROMS3SETTINGS_SCHEME._serialized_end=922
  _IMPORTFROMS3RESULT._serialized_start=924
  _IMPORTFROMS3RESULT._serialized_end=944
  _IMPORTFROMS3METADATA._serialized_start=947
  _IMPORTFROMS3METADATA._serialized_end=1132
  _IMPORTFROMS3REQUEST._serialized_start=1135
  _IMPORTFROMS3REQUEST._serialized_end=1273
  _IMPORTFROMS3RESPONSE._serialized_start=1275
  _IMPORTFROMS3RESPONSE._serialized_end=1343
  _YDBDUMPFORMAT._serialized_start=1345
  _YDBDUMPFORMAT._serialized_end=1377
  _IMPORTDATARESULT._serialized_start=1379
  _IMPORTDATARESULT._serialized_end=1397
  _IMPORTDATAREQUEST._serialized_start=1400
  _IMPORTDATAREQUEST._serialized_end=1574
  _IMPORTDATARESPONSE._serialized_start=1576
  _IMPORTDATARESPONSE._serialized_end=1642
# @@protoc_insertion_point(module_scope)
